/*
 * RGB-Matrix-Uhr 64x32, Scott-Falk Hhn, Matrix-Uhr.c, v2.20, 16.02.2024
 * AtmelStudio Editor-Einstellung: Tools / Options / Text Editor / GCC / Tabs / Smart, Tab size: 2, Indent size: 2, Insert spaces
 * Einige Ideen und Tricks stammen aus der RGBmatrixPanel-Library fr Arduino von Adafruit, vielen Dank an die Entwickler.
 *
 * Taktfrequenz 20,0 MHz
 *
 * Belegung der I/O-Ports:
 * PB0 - LED-Matrix - Ausgang CLK
 * PA0 - LED-Matrix - Ausgang A
 * PA1 - LED-Matrix - Ausgang B
 * PA2 - LED-Matrix - Ausgang C
 * PA3 - LED-Matrix - Ausgang D, bei Matrix-Scan-Modus 1/8 nicht genutzt
 * PA4 - LED-Matrix - Ausgang LAT
 * PA5 - LED-Matrix - Ausgang OE
 * PC2 - LED-Matrix - Ausgang R1
 * PC3 - LED-Matrix - Ausgang G1
 * PC4 - LED-Matrix - Ausgang B1
 * PC5 - LED-Matrix - Ausgang R2
 * PC6 - LED-Matrix - Ausgang G2
 * PC7 - LED-Matrix - Ausgang B2
 * PA6 - DCF77-Empfnger - Eingang, Pull-up schaltbar
 * PA7 - Helligkeit - Eingang analog
 * PD0 - Timing-Jumper - Eingang mit Pull-up
 * PD2 - RX-Datenleitung vom RasPi - Eingang mit Pull-up
 * PD3 - TX-Datenleitung zum RasPi - Ausgang
 * PD4 - Test-LED - Ausgang
 * PD5 - Taster 3 - Eingang mit Pull-up
 * PD6 - Taster 2 - Eingang mit Pull-up
 * PD7 - Taster 1 - Eingang mit Pull-up
 * PD1 - Sound-Modul - Ausgang Enable (Low-aktiv)
 * PB1 - Sound-Modul - Ausgang 1
 * PB2 - Sound-Modul - Ausgang 2
 * PB3 - Sound-Modul - Ausgang 3
 * PB4 - Sound-Modul - Ausgang 4
 * PB5 - Sound-Modul - Ausgang 5
 * PB6 - Sound-Modul - Ausgang 6
 *
 * BODLEVEL  = 4V3
 * OCDEN     = [ ]
 * JTAGEN    = [ ]
 * SPIEN     = [X]
 * WDTON     = [ ]
 * EESAVE    = [X]
 * BOOTSZ    = 4096W_F000
 * BOOTRST   = [ ]
 * CKDIV8    = [ ]
 * CKOUT     = [ ]
 * SUT_CKSEL = FSOSC_16KCK_65MS_XOSC_SLOWPWR
 *
 * EXTENDED  = 0xFC
 * HIGH      = 0xD1
 * LOW       = 0xF7
 *
 * Der gesamte Matrix-Puffer besteht aus 4096 Bytes, die in 4 Blcke unterteilt sind:
 *
 * Block 1: 0000-1023 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 1 (Bit 2-4)  \
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 1 (Bit 5-7)   \ Matrix 0
 * Block 2: 1024-2047 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 2 (Bit 2-4)   /
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 2 (Bit 5-7)  /
 * Block 3: 2048-3071 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 1 (Bit 2-4)  \
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 1 (Bit 5-7)   \ Matrix 1
 * Block 4: 3072-4095 - Pixel   0- 511 (obere Matrix-Hlfte)  - Farbwert 2 (Bit 2-4)   /
 *                      Pixel 512-1023 (untere Matrix-Hlfte) - Farbwert 2 (Bit 5-7)  /
 *
 * Es ist immer nur Matrix 0 oder Matrix 1 aktiv. Alle darzustellenden Daten werden in die inaktive Matrix geschrieben, whrend die Interrupt-Routine von Timer 1 die Daten der aktiven Matrix ausgibt.
 * Somit erfolgt der Bildaufbau im Hintergrund und erst nach Fertigstellung des kompletten Bildes wird durch eine Umschaltung der Matrix der neue Bildinhalt sichtbar.
 *
 * Jedes Byte der Matrix ist gleich aufgebaut und die Daten sind so angeordnet, dass ein Farbbyte direkt mit hoher Geschwindigkeit auf die Matrix ausgegeben werden kann:
 *
 * Bit 0: 0  - nicht genutzt  \ An diesen Ports (PC0 und PC1) befindet sich der TWI-Bus (I2C) des Controllers.
 * Bit 1: 0  - nicht genutzt  / Die Portbelegung wurde von Adafruit bernommen, obwohl TWI nicht genutzt wird.
 * Bit 2: R1 - Rot-Wert von Pixel n (obere Bildhlfte)
 * Bit 3: G1 - Grn-Wert von Pixel n (obere Bildhlfte)
 * Bit 4: B1 - Blau-Wert von Pixel n (obere Bildhlfte)
 * Bit 5: R2 - Rot-Wert von Pixel n + 1024 (untere Bildhlfte)
 * Bit 6: G2 - Grn-Wert von Pixel n + 1024 (untere Bildhlfte)
 * Bit 7: B2 - Blau-Wert von Pixel n + 1024 (untere Bildhlfte)
 */

#define F_CPU 20000000UL
#include <avr/io.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>

const char strd00[] PROGMEM = "V2.20 * SFH";

const uint8_t c_yearh  = 20;      // Jahrhundert festlegen: 20xx
const uint8_t c_year   = 24;      // frhestes Jahr festlegen (fr manuelle Zeiteingabe)
const uint8_t c_start  = 10;      // Startzeit in Sekunden, whrend dieser Zeit werden Versionsinformationen angezeigt
const uint8_t c_setup  = 120;     // Anzeige der Einstellungs-Mens in Sekunden, nach Ablauf ohne Eingabe wird zur normalen Anzeige zurckgekehrt (maximal 254)
const uint8_t c_sync   = 25;      // Zeit in Stunden bis zur Meldung eines Sync-Ausfalls (maximal 254)
const uint8_t c_dcf    = 30;      // Zeit in Minuten bis zur Wiedereinschaltung der LED-Matrix bei eingeschalteter DCF-Synchronisierung (maximal 254)
const uint8_t c_keyl1  = 20;      // Zeit bis zum Erkennen eines langen Tastendrucks in Einheiten von 40 ms
const uint8_t c_keyl2  = 3;       // Wiederholungszeit fr Auto-Repeat in Einheiten von 40 ms (muss immer kleiner sein als c_keyl1)
const uint8_t c_keyl3  = 5;       // Anzahl der Zyklen von langen Tastendrcken in Einheiten von c_keyl1 (wird fr Poweroff-Erkennung verwendet)
const uint8_t c_setout = 120;     // Speicherzeit von Sensor- und Geburtstagsdaten in Sekunden (maximal 254)
const uint8_t c_altout = 120;     // Speicherzeit von Alarmdaten in Sekunden (maximal 254)
const uint8_t c_sndto1 = 25;      // Zeitdauer des Impulses fr die Soundausgabe in Einheiten von 40ms (maximal 254)
const uint8_t c_sndto2 = 150;     // Gesamtdauer der Wartezeit bis zur nchsten Soundausgabe in Einheiten von 40ms (muss immer grer sein als c_sndto1, maximal 254)
const uint8_t c_sndena = 10;      // Zeit in Sekunden bis zum Aktivieren des Soundmoduls nach Systemstart (maximal 254)
const uint8_t c_mute   = 12;      // Zeit in Stunden bis zur automatischen Abschaltung der Mute-Funktion (maximal 254)
const uint8_t c_natout = 30;      // Zeit in Sekunden fr die vorbergehende Aktivierung des Displays zur Nachtzeit (maximal 254)

// EEPROM-Adressen (werden als Konstanten definiert, um eine feste Adresse zu erhalten)

const uint16_t e_lang   =   0;    // feste EEPROM-Adresse 0000 - Sprache (siehe lang)
const uint16_t e_dcfsyn =   1;    // feste EEPROM-Adresse 0001 - Feste DCF-Synchronisierungszeit (siehe dcfsyn)
const uint16_t e_dcfinv =   2;    // feste EEPROM-Adresse 0002 - DCF-Invertierung (siehe dcfinv)
const uint16_t e_dcfpup =   3;    // feste EEPROM-Adresse 0003 - DCF-Pull-Up (siehe dcfpup)
const uint16_t e_tstled =   4;    // feste EEPROM-Adresse 0004 - Test-LED (siehe tstled)
const uint16_t e_colclk =   5;    // feste EEPROM-Adresse 0005 - Farbe fr Uhrzeit (siehe colclk)
const uint16_t e_coldat =   6;    // feste EEPROM-Adresse 0006 - Farbe fr Datum, Wochentag und Sensordaten (siehe coldat)
const uint16_t e_colalm =   7;    // feste EEPROM-Adresse 0007 - Farbe fr Alarmmeldungen (siehe colalm)
const uint16_t e_brimin =   8;    // feste EEPROM-Adresse 0008 - Minimale Helligkeit (siehe brimin)
const uint16_t e_brimax =   9;    // feste EEPROM-Adresse 0009 - Maximale Helligkeit (siehe brimax)
const uint16_t e_senfil =  10;    // feste EEPROM-Adresse 0010 - Sensorfilter (siehe senfil)
const uint16_t e_alarms =  11;    // feste EEPROM-Adresse 0011 - Alarme (siehe alarms)
const uint16_t e_sennum =  12;    // feste EEPROM-Adresse 0012-0019 - Sensornummern 1-8 (siehe sennum[])
const uint16_t e_sennam =  20;    // feste EEPROM-Adresse 0020-0059 - Sensornamen 1-8 (siehe sennam[])
const uint16_t e_almnor =  60;    // feste EEPROM-Adresse 0060-0063 - Normalzustnde der Alarme 1-4 (siehe almnor[])
const uint16_t e_almnam =  64;    // feste EERPOM-Adresse 0064-0111 - Namen der Alarme 1-4 (siehe almnam[])
const uint16_t e_sounfr = 112;    // feste EERPOM-Adresse 0112 - Sound von (siehe sounfr)
const uint16_t e_sounto = 113;    // feste EERPOM-Adresse 0113 - Sound bis (siehe sounto)
const uint16_t e_housnd = 114;    // feste EERPOM-Adresse 0114 - Stundengong (siehe housnd)
const uint16_t e_almsnd = 115;    // feste EERPOM-Adresse 0115 - Alarmsound (siehe almsnd)
const uint16_t e_msgscr = 116;    // feste EEPROM-Adresse 0116 - Roll-Geschwindigkeit (siehe msgscr)
const uint16_t e_avocol = 117;    // feste EEPROM-Adresse 0117 - Farbe meiden (siehe avocol)
const uint16_t e_shdate = 118;    // feste EEPROM-Adresse 0118 - Datum anzeigen (siehe shdate)
const uint16_t e_shwday = 119;    // feste EEPROM-Adresse 0119 - Wochentag anzeigen (siehe shwday)
const uint16_t e_chtime = 120;    // feste EEPROM-Adresse 0120 - Wechselzeit der Anzeige (siehe chtime)
const uint16_t e_autost = 121;    // feste EEPROM-Adresse 0121 - Automatische Sommerzeit (siehe autost)
const uint16_t e_timezo = 122;    // feste EEPROM-Adresse 0122 - Zeitzone (siehe timezone)
const uint16_t e_daymod = 123;    // feste EEPROM-Adresse 0123 - Taganzeige-Modus (siehe daymod)
const uint16_t e_dcfhou = 124;    // feste EEPROM-Adresse 0124 - DCF-Synchronisierungsstunde (siehe dcfhou)
const uint16_t e_nighfr = 125;    // feste EERPOM-Adresse 0125 - Nacht von (siehe nighfr)
const uint16_t e_nighto = 126;    // feste EERPOM-Adresse 0126 - Nacht bis (siehe nighto)

// Folgende Variablen werden beim Systemstart aus dem EEPROM gelesen

static uint8_t lang;              // Sprache (0-1), 0 = deutsch, 1 = englisch
static uint8_t dcfsyn;            // Feste DCF77-Synchronisierungszeit (0-1), 1 = Abschaltung der LED-Matrix zur sicheren DCF-Synchronisierung
static uint8_t dcfhou;            // DCF77-Synchronisierungsstunde (0-23), Beginn einer DCF-Synchronisierung mit Display-Abschaltung
static uint8_t dcfinv;            // DCF77-Invertierung (0-1), 0  = DCF-Empfnger liefert Low-Impulse, 1 = DCF-Empfnger liefert High-Impulse
static uint8_t dcfpup;            // DCF77-Pull-up (0-1), 0 = Pull-up ausgeschaltet, 1 = Pull-up eingeschaltet
static uint8_t tstled;            // Test-LED (0-2), 0 = Test-LED ausgeschaltet, 1 = Anzeige DCF77-Signal, 2 = Anzeige RS-232-Signal
static uint8_t colclk;            // Farbe fr Uhrzeit (0-16), 0 = Zufall, 1-16 Farbe aus Tabelle
static uint8_t coldat;            // Farbe fr Datum, Wochentag und Sensordaten (0-16), 0 = Zufall, 1-16 Farbe aus Tabelle
static uint8_t colalm;            // Farbe fr Alarmmeldungen (0-16), 0 = Zufall,  1-16 Farbe aus Tabelle
static uint8_t avocol;            // Farbe meiden (0-16), 0 = aus, 1-16 Farbe aus Tabelle
static uint8_t brimin;            // Minimale Helligkeit der LED-Matrix (0-9)
static uint8_t brimax;            // Maximale Helligkeit der LED-Matrix (0-9)
static uint8_t chtime;            // Wechselzeit der Anzeige in Sekunden (2-4)
static uint8_t shdate;            // Datum anzeigen (0-1), 0 = Datum aus, 1 = Datum ein
static uint8_t shwday;            // Wochentag anzeigen (0-1), 0 = Wochentag aus, 1 = Wochentag ein
static uint8_t daymod;            // Taganzeige-Modus (0-2), 0 = Kalendertag, 1 = Wochentag, 2 = Wechsel
static uint8_t nighfr;            // Nacht von (0-23), ab dieser Stunde wird die LED-Matrix in den Nachtmodus geschaltet
static uint8_t nighto;            // Nacht bis (0-23), bis zu dieser Stunde wird die LED-Matrix in den Nachtmodus geschaltet
static uint8_t autost;            // Automatische Sommerzeit (0-1), 0 = aus, 1 = ein
static uint8_t timezone;          // Zeitzone (0-1), 0 = MEZ, 1 = MESZ
static uint8_t senfil;            // Sensorfilter (0-1), 0 = Filter aus - alle Sensoren anzeigen, 1 = Filter ein - nur konfigurierte Sensoren 1-8 anzeigen
static uint8_t alarms;            // Alarme (0-1), 1 = Alarme vom Sensormodul werden angezeigt, beim Wechsel auf 1 wird der aktuelle Alarmzustand als Normalzustand gespeichert
static uint8_t sennum[8];         // Sensornummern 1-8 (0-31), 0 = Sensor nicht genutzt, 1-31 = Sensornummer des Temperatur-Messsystems (nur relevant fr Datenempfang vom Sensormodul)
static uint8_t sennam[8][5];      // Sensornamen 1-8 mit jeweils 5 Zeichen (nur relevant fr Datenempfang vom Sensormodul)
static uint8_t almnor[4];         // Ruhezustnde der Alarme 1-4 vom Sensormodul (0-1) (nur relevant fr Datenempfang vom Sensormodul)
static uint8_t almnam[4][12];     // Namen der Alarme 1-4 vom Sensormodul mit jeweils 12 Zeichen (nur relevant fr Datenempfang vom Sensormodul)
static uint8_t sounfr;            // Sound von (0-23), ab dieser Stunde werden akustische Alarme ausgegeben
static uint8_t sounto;            // Sound bis (0-23), bis zu dieser Stunde werden akustische Alarme ausgegeben
static uint8_t housnd;            // Stundengong (0-1), 1 = akustisches Stundensignal eingeschaltet
static uint8_t almsnd;            // Alarmsound (0-1), 1 = akustisches Alarmsignal eingeschaltet
static uint8_t msgscr;            // Roll-Geschwindigkeit der Textnachricht (0-20)

// Weitere Variablen

static uint8_t matrix[2][2048];   // Matrix-Puffer, Aufbau siehe weiter oben
static uint8_t matnum;            // Aktuelle Matrix (0-1)
static uint8_t matscn;            // Matrix-Scan-Modus (0-1), 0 = 1/16, 1 = 1/8
static uint8_t mulcnt;            // Multiplex-Zhler, wird alle 15,625 s weitergezhlt und fr die Datenausgabe an die LED-Matrix bentigt
static uint8_t bright[32];        // Helligkeits-Puffer, enthlt die letzten 32 im Abstand von 20ms gelesenen ADC-Werte vom Helligkeitssensor
static uint8_t bripos;            // aktuelle Schreibposition des Helligkeits-Puffers
static uint8_t bripwm;            // Helligkeitswert fr PWM-Steuerung der LED-Matrix
static uint8_t color1;            // RGB-Farb-Komponente 1 (3 Bits)  \ wird spter bei der Multiplex-Ausgabe abwechselnd
static uint8_t color2;            // RGB-Farb-Komponente 2 (3 Bits)  / ausgegeben, um 16 Farbwerte darstellen zu knnen
static uint8_t posx;              // X-Position fr die Zeichenausgabe (0-63)
static uint8_t posy;              // Y-Position fr die Zeichenausgabe (0-31)
static uint8_t undrline;          // Unterstreichung (0-1), 1 = Unterstreichung eingeschaltet
static uint8_t colclkr;           // Zufallsfarbe fr Uhrzeit (0-15)
static uint8_t coldatr;           // Zufallsfarbe fr Datum, Wochentag und Sensordaten (0-15)
static uint8_t colalmr;           // Zufallsfarbe fr Alarmmeldungen (0-15)
static uint8_t dcfstat1;          // DCF77, Signal-Status beim vorherigen Interrupt (0-1)
static uint8_t dcfstat2;          // DCF77, Signalisierung des Minutenwechsels (0-1)
static uint8_t dcfcnt;            // DCF77-Zhler zur Ermittlung der Impulslnge
static uint8_t dcfimp;            // DCF77-Zhler zur Ermittlung der Impulsanzahl
static uint8_t key1old;           // Taster 1, Zustand bei der letzten Abfrage
static uint8_t key1stat;          // Taster 1, aktueller stabiler Zustand (0-1)
static uint8_t key1quit;          // Taster 1, Quittierungs-Status (0-1)
static uint8_t key1long;          // Taster 1, Zhler fr langen Tastendruck (0-c_keyl1)
static uint8_t key2old;           // Taster 2, Zustand bei der letzten Abfrage
static uint8_t key2stat;          // Taster 2, aktueller stabiler Zustand (0-1)
static uint8_t key2quit;          // Taster 2, Quittierungs-Status (0-1)
static uint8_t key2long;          // Taster 2, Zhler fr langen Tastendruck (0-c_keyl1)
static uint8_t key2lcnt;          // Taster 2, Zykluszhler fr langen Tastendruck (0-c_keyl3)
static uint8_t key3old;           // Taster 3, Zustand bei der letzten Abfrage
static uint8_t key3stat;          // Taster 3, aktueller stabiler Zustand (0-1)
static uint8_t key3quit;          // Taster 3, Quittierungs-Status (0-1)
static uint8_t dispmode;          // Anzeige-Modus (0-4)
                                  // 0 = Normaler Modus (Zeit in groen Ziffern; Untere Zeile: Datum, Wochentag und Sensordaten im Wechsel)
                                  // 1 = Alarm-Modus (Zeit verkleinert; Mittlere Zeile: Datum, Wochentag und Sensordaten im Wechsel; Untere Zeile: Alarmmeldungen im Wechsel)
                                  // 2 = Nachrichten-Modus (Zeit verkleinert; Mittlere Zeile: Datum, Wochentag und Sensordaten im Wechsel; Untere Zeile: Nachricht als Laufschrift
                                  // 3 = Systemstart-Modus (Zeit verkleinert; Mittlere und untere Zeile: Versions-Informationen)
                                  // 4 = Einstellungs-Modus (Zeit verkleinert; Mittlere Zeile: Einstellung; Untere Zeile: Parameter)
static uint8_t setmode;           // Einstellungs-Modus (0-47)
                                  //  0 = Farbe fr Uhrzeit                  1 = Farbe fr Datum, Wochentag         2 = Farbe fr Alarmmeldungen           3 = Farbe meiden
                                  //  4 = Helligkeit Minimal                 5 = Helligkeit Maximal                 6 = Test-LED                           7 = Feste DCF-Sync.-Zeit
                                  //  8 = DCF-Invertierung                   9 = DCF-Pull-up                       10 = Stundengong                       11 = Alarmsignal
                                  // 12 = Sound von                         13 = Sound bis                         14 = Roll-Geschwindigkeit              15 = Sensorfilter
                                  // 16 = Sensornummer 1                    17 = Sensornummer 2                    18 = Sensornummer 3                    19 = Sensornummer 4
                                  // 20 = Sensornummer 5                    21 = Sensornummer 6                    22 = Sensornummer 7                    23 = Sensornummer 8
                                  // 24 = Sensorname 1                      25 = Sensorname 2                      26 = Sensorname 3                      25 = Sensorname 4
                                  // 28 = Sensorname 5                      29 = Sensorname 6                      30 = Sensorname 7                      31 = Sensorname 8
                                  // 32 = Alarme                            33 = Alarmname 1                       34 = Alarmname 2                       35 = Alarmname 3
                                  // 36 = Alarmname 4                       37 = Tag                               38 = Monat                             39 = Jahr
                                  // 40 = Stunde                            41 = Minute                            42 = Sekunde                           43 = Matrix-Test
                                  // 44 = Sound-Test                        45 = Sprache                           46 = Datum anzeigen                    47 = Wochentag anzeigen
                                  // 48 = Wechselzeit                       49 = Automatische Sommerzeit           50 = Taganzeige                        51 = DCF-Synchronisierungsstunde
                                  // 52 = Nacht von                         53 = Nacht bis
static uint8_t setcount;          // Einstellungszhler (0-53), ber diesen Zhler und modelist[] erfolgt der Zugriff auf den Einstellungs-Modus setmode
static uint8_t editmode;          // Editier-Modus fr die Eingabe von Sensor- und Alarmnamen (0-1):
                                  // 0 = Taster 1 fhrt zur nchsten Einstellung
                                  // 1 = Taster 1 fhrt zum nchsten Zeichen bei der Eingabe der Sensor- und Alarmnamen
static uint8_t editpos;           // Zeichenposition, die im Editier-Modus gerade bearbeitet wird (0-11)
static uint8_t nighttime;         // Nachtzeit fr Display (0-1)
static uint8_t intcnt;            // Interruptzhler (0-49)
static uint8_t second;            // Sekundenzhler (0-59)
static uint8_t minute;            // Minutenzhler (0-59)
static uint8_t hour;              // Stundenzhler (0-23)
static uint8_t wday;              // Wochentagzhler (1-7)
static uint8_t day;               // Tageszhler (1-31)
static uint8_t month;             // Monatszhler (1-12)
static uint8_t year;              // Jahres-Zhler (0-99)
static uint8_t alm_flag;          // Alarm-Flag, mindestens 1 Alarm ist aktiv
static uint8_t msg_flag;          // Nachrichten-Flag, eine Textnachricht wurde empfangen
static uint8_t lasthour;          // Letzte Stunde, an der ein automatischer Farbwechsel durchgefhrt wurde
static int8_t datacoun;           // Zhler fr die auszugebenden Datenfelder, zhlt alle 2-4 Sekunden von 0 aufwrts:
                                  // 0    = Ausgabe des Datums (wird bersprungen, wenn "Datum anzeigen" ausgeschaltet ist)
                                  // 1    = Ausgabe des Wochentages (wird bersprungen, wenn "Wochentag anzeigen" ausgeschaltet ist)
                                  // 2-32 = Ausgabe eines Sensorwertes oder Geburtstages (leere Felder werden bersprungen), mit Sensorfilter: Bereich 2-9
                                  // 33   = Ausgabe von "keine Daten" (nur wenn Datum und Wochentag ausgeschaltet sind und keine Sensordaten vorhanden sind)
static uint8_t almcoun;           // Zhler fr die auszugebenden Alarmfelder, zhlt alle 2-4 Sekunden von 0 aufwrts
static uint8_t daymcoun;          // Zhler fr die Taganzeige, wechselt alle 2-4 Sekunden zwischen 0 und 1
static uint8_t syncstat;          // Status der Synchronsierung (0-4)
                                  // 0 = Uhr ist nicht synchronisert (Status nach Systemstart) -> Doppelpunkt in Uhrzeit erscheint in der Alarmfarbe
                                  // 1 = Uhr wurde manuell gestellt
                                  // 2 = Uhr wurde mindenstens 24 Stunden nicht synchronisiert (Sync-Ausfall) -> Doppelpunkt in Uhrzeit erscheint in der Datumfarbe
                                  // 3 = Uhr wurde ber DCF77 synchronisiert
                                  // 4 = Uhr wurde ber das Web-Modul synchronisiert
static uint8_t disptout;          // Anzeige-Timeout-Zhler, nach Ablauf der Zeit wird zum normalen Anzeige-Modus zurckgekehrt; wird vom Hauptprogramm gesetzt und vom Timer-Interrupt im Sekundentakt
                                  // auf 0 gezhlt, das Hauptprogramm setzt nach Bearbeitung den Wert auf 255 (Stopp)
static uint8_t synctout;          // Synchronisierungs-Timeout-Zhler, nach Ablauf der Zeit wird der Status auf "Sync-Ausfall" gesetzt; wird vom Hauptprogramm gesetzt und vom Timer-Interrupt im
                                  // Stundentakt auf 0 gezhlt, das Hauptprogramm setzt nach Bearbeitung den Wert auf 255 (Stopp)
static uint8_t mutetout;          // Stummschaltungs-Timeout-Zhler, nach Ablauf der Zeit wird die Stummschaltung wieder aufgehoben; wird vom Hauptprogramm gesetzt und vom Timer-Interrupt im
                                  // Stundentakt auf 0 gezhlt, das Hauptprogramm setzt nach Bearbeitung den Wert auf 255 (Stopp)
static uint8_t dcftout;           // DCF-Timeout-Zhler, wird bei eingeschalteter fester DCF-Synchronisierungszeit (dcfsyn = 1) kurz nach dem Systemstart und zur in dcfhou festgelegten Stunde
                                  // gestartet; whrend der Laufzeit wird die LED-Matrix ausgeschaltet; nach erfolgreicher Synchronisierung oder nach Ablauf der Zeit in c_dcf wird der Zhler gestoppt
                                  // und die LED-Matrix wieder eingeschaltet
static uint8_t nacttout;          // Nachtaktivierungs-Timeout-Zhler, whrend der Nacht wird das Display durch den Quittierungssensor aktiviert und nach Ablauf der Zeit wieder deaktiviert; wird vom
                                  // Hauptprogramm gesetzt und im Sekundentakt auf 0 gezhlt, das Hauptprogramm setzt nach Bearbeitung den Wert auf 255 (Stopp)
static uint8_t tim0cnt1;          // Vorteiler 1 fr Timer 0, 625 Hz -> 25 Hz
static uint8_t tim0cnt2;          // Vorteiler 2 fr Timer 0, 25 Hz -> 1 Hz
static uint8_t tim0sec;           // Sekundenzhler fr Timer 0
static uint8_t tim0min;           // Minutenzhler fr Timer 0
static char strbuf[30];           // Stringpuffer fr die LED-Textausgabe und die Generierung von Eintrgen fr das Sensor- und Alarmfeld
static uint8_t dcftab[42];        // DCF77-Tabelle, enthlt empfangene Bits, wobei jedes Bit ein Byte der Tabelle belegt. Eingehende Bits werden in das letzte Byte eingeschoben und alle weiteren
                                  // Bytes um eine Position weitergerckt; die Bytes enthalten nach einem kompletten Minutenzyklus folgende Daten:
                                  // Byte 00:    Zeitzone, 0 = Winterzeit, 1 = Sommerzeit
                                  // Byte 01:    Zeitzone, 1 = Winterzeit, 0 = Sommerzeit
                                  // Byte 02:    Schaltsekunde (nicht ausgewertet)
                                  // Byte 03:    Startbit (nicht ausgewertet)
                                  // Byte 04-07: Minute Einer
                                  // Byte 08-10: Minute Zehner
                                  // Byte 11:    Prfbit 1 (Minuten)
                                  // Byte 12-15: Stunde Einer
                                  // Byte 16-17: Stunde Zehner
                                  // Byte 18:    Prfbit 2 (Stunden)
                                  // Byte 19-22: Kalendertag Einer
                                  // Byte 23-24: Kalendertag Zehner
                                  // Byte 25-27: Wochentag
                                  // Byte 28-31: Monat Einer
                                  // Byte 32:    Monat Zehner
                                  // Byte 33-36: Jahr Einer
                                  // Byte 37-40: Jahr Zehner
                                  // Byte 41:    Prfbit 3 (Datum)
static uint8_t dcfbuf1[7];        // DCF-Puffer 1, aktuelle dekodierte Uhrzeit:
                                  // Byte 0: Zeitzone (0-1), 0 = Winterzeit, 1 = Sommerzeit
                                  // Byte 1: Minute (0-59)
                                  // Byte 2: Stunde (0-23)
                                  // Byte 3: Kalendertag (1-31)
                                  // Byte 4: Wochentag (1-7)
                                  // Byte 5: Monat (1-12)
                                  // Byte 6: Jahr (0-99)
static uint8_t dcfbuf2[7];        // DCF-Puffer 2, dekodierte Uhrzeit eine Minute zuvor (korrigiert um eine Minute), gleicher Aufbau wie dcbuf1
static uint8_t dcfbuf3[7];        // DCF-Puffer 3, dekodierte Uhrzeit eine Minute zuvor (korrigiert um eine Minute), gleicher Aufbau wie dcbuf1
static uint8_t rxbuff[256];       // USART-Empfangspuffer
static char rxdata[100];          // USART-Empfangsdaten
static uint8_t rxwrite;           // Schreib-Zeiger fr USART-Empfangspuffer
static uint8_t rxread;            // Lese-Zeiger fr USART-Empfangspuffer
static uint8_t rxpos;             // USART-Empfangsdaten-Position
static char senarray[31][12];     // Sensorfeld mit 31 Pltzen und jeweils 12 Zeichen fr Sensor- und Geburtstagsdaten
static uint8_t sentout[31];       // Sensorfeld-Timeout-Zhler mit 31 Pltzen, wird beim Schreiben eines neuen Datensatzes auf [c_setout / chtime] gesetzt und im Wechselzeit-Takt rckwrts gezhlt,
                                  // bei 0 wird der zugehrige Datensatz gelscht und der Zhler auf 255 gesetzt (Stopp)
static uint8_t almsold[4];        // vorherige Zustnde (0-1) der Alarme 1-4 vom Temperatur-Messsystem (wie vom Sensormodul empfangen)
static uint8_t almsnew[4];        // aktuelle Zustnde (0-1) der Alarme 1-4 vom Temperatur-Messsystem (wie vom Sensormodul empfangen)
static char almarray[8][12];      // Alarmfeld mit 8 Pltzen und jeweils 12 Zeichen fr Alarme, die Pltze 0-3 werden immer von den Alarmen 1-4 des Sensormoduls verwendet
static uint8_t almtout[8];        // Alarmfeld-Timeout-Zhler mit 8 Pltzen, wird bei aktivem Alarm auf [c_altout / chtime] gesetzt und im Wechselzeit-Takt rckwrts gezhlt, bei 0 wird der
                                  // zugehrige Datensatz gelscht bzw. markiert und der Zhler auf 255 gesetzt (Stopp)
static uint8_t almstat[8];        // Alarmfeld-Status mit 8 Pltzen (0-1), 1 = Alarm aktiv
static uint8_t almhold[8];        // Alarmfeld-Halte-Flag mit 8 Pltzen (0-1), 1 = Alarm wird nicht automatisch nach Rckkehr in den Normalzustand gelscht (Quittierung erforderlich)
static uint8_t sndflags[6];       // Sound-Flags fr die Ausgabe von akustischen Alarm- und Stundensignalen, 1 = Signal steht an
                                  // Byte 0: Alarmsignal 1 (ist dem Alarm 1 vom Sensormodul fest zugeordnet, kann bei Steuerung ber das Web-Modul frei verwendet werden)
                                  // Byte 1: Alarmsignal 2 (ist dem Alarm 2 vom Sensormodul fest zugeordnet, kann bei Steuerung ber das Web-Modul frei verwendet werden)
                                  // Byte 2: Alarmsignal 3 (ist dem Alarm 3 vom Sensormodul fest zugeordnet, kann bei Steuerung ber das Web-Modul frei verwendet werden)
                                  // Byte 3: Alarmsignal 4 (ist dem Alarm 4 vom Sensormodul fest zugeordnet, kann bei Steuerung ber das Web-Modul frei verwendet werden)
                                  // Byte 4: Alarmsignal 5 (ist der Textnachricht fest zugeordnet, kann aber auch fr Alarme verwendet werden)
                                  // Byte 5: Alarmsignal 6 (ist dem Stundensignal fest zugeordnet, kann aber auch fr Alarme verwendet werden)
static uint8_t actsound;          // enthlt die aktuelle ausgegebene Soundnummer (0-5), 255 = Sound inaktiv
static uint8_t sndtout;           // Sound-Timeout-Zhler, zhlt bei Soundausgabe von 0 bis c_sndto1 (Soundausgang abschalten), zhlt weiter bis c_sndto2 (Wartezeitende) und wird dann auf 255 gesetzt
static uint8_t sounflag;          // Sound-Flag, wird im Zeitraum "Sound von" - "Sound bis" gesetzt und bei der Erzeugung der akustischen Alarme ausgewertet
static uint8_t sndtest;           // Sound-Test-Zhler (0-5), wird im Einstellungs-Men beim Sound-Test verwendet, zhlt bei jedem Test weiter
static uint8_t sndena;            // Timeout-Zhler fr das verzgerte Aktivieren des Sound-Moduls (verhindert Gerusche beim Programmieren des Mikrocontrollers)
static uint8_t mute;              // manuelle Stummschaltung (0-1), 1 = keine akustischen Alarme ausgeben
static uint8_t quitcnt;           // Quittierungs-Zhler zur Anzeige der Animation beim Quittieren (0-32)
static uint8_t msgpbuf[1000];     // Pixelpuffer fr die Textnachricht, der Platz reicht fr 500 Pixelspalten, die jeweils 2 Bytes bentigen
static int16_t msgpbend;          // zeigt auf das Ende der Nachricht im Pixelpuffer
static int16_t msgpbpos;          // zeigt auf die aktuelle Pixelspalte im Pixelpuffer, wird fr die Laufschrift-Ausgabe bentigt
static uint8_t msgscrol;          // Nachrichten-Roll-Timeout-Zhler, zhlt alle 1,6ms bis zum einem von msgscr festgelegten Wert, nach Ablauf wird der Zhler zurckgesetzt und msgpbpos auf die
                                  // nchste Pixelzeile gesetzt (vom Timer-Interrupt gesteuert)
static uint8_t msgstest;          // zhlt fortlaufend rckwrts im gleichen Takt wie msgpbpos, wird fr das Einstellung-Men von msgscr bentigt
static uint8_t matrtest;          // LED-Matrix-Test-Modus (0-1), 1 = Test aktiv (alle LEDs leuchten wei)
static uint8_t tstlcnt;           // Test-LED-Zhler (0-10), wird bei jedem empfangenen RS-232-Paket auf 0 gesetzt und alle 20ms bis 10 hochgezhlt, whrend der Zhlzeit (200ms) leuchtet die Test-LED
static uint8_t saveflag;          // Speicher-Flag (0-1), wird gesetzt, wenn Einstellungen im EEPROM gespeichert werden mssen

extern void matrix_out(uint16_t *, uint8_t, uint8_t, uint8_t);   // externe Assembler-Routine zur Datenausgabe an die LED-Matrix (Parameter: Bitmap-Zeiger, Multiplex-Zhler, PWM-Wert und Scan-Modus)

#include "tables.h"               // Verschiedene Tabellen und Listen
#include "strings.h"              // Verschiedene Texte in Deutsch und Englisch
#include "charset.h"              // Zeichensatz-Definitionen
#include "interrupt.h"            // Interrupt-Routinen
#include "functions.h"            // Verschiedene Funktionen

// --- Beginn Initialisierung ---

int main(void)
{
  int8_t i, j;                                                                                        // Zhlervariablen
  PORTA  = _BV(PINA5) | _BV(PINA3);                                                                   // PortA: Matrix-LAT auf Low, Matrix-OE auf High, Pull-up an Matrix-D
  DDRA   = _BV(PINA5) | _BV(PINA4) | _BV(PINA2) | _BV(PINA1) | _BV(PINA0);                            // PortA: Matrix-A, B und C sowie LAT und OE auf Ausgang, Matrix-D bleibt zunchst auf Eingang
  PORTB  = 0;                                                                                         // PortB: alles auf Low
  DDRB   = _BV(PINB6) | _BV(PINB5) | _BV(PINB4) | _BV(PINB3) | _BV(PINB2) | _BV(PINB1) | _BV(PINB0);  // PortB: Pins 0-6 auf Ausgang
  PORTC  = 0;                                                                                         // PortC: alles auf Low
  DDRC   = _BV(PINC7) | _BV(PINC6) | _BV(PINC5) | _BV(PINC4) | _BV(PINC3) | _BV(PINC2);               // PortC: Pins 2-7 auf Ausgang
  PORTD  = _BV(PIND7) | _BV(PIND6) | _BV(PIND5) | _BV(PIND4) | _BV(PIND2) | _BV(PIND1) | _BV(PIND0);  // PortD: Pin 0, 2 und 5-7 Pull-up, Pin 4 und 1 auf High
  DDRD   = _BV(PIND4) | _BV(PIND1);                                                                   // PortD: Pin 4 und 1 auf Ausgang
  TCCR0A = _BV(WGM01);                                                                                // Timer0: wird fr Timeout-Zhler verwendet
  TCCR0B = _BV(CS02);                                                                                 // Timer0: Mode 2 (CTC), Vorteiler = 256
  OCR0A  = F_CPU / 256 / 625 - 1;                                                                     // Timer0: Endwert fr 625Hz setzen (einzig mglicher ganzzahliger Wert)
  TIMSK0 = _BV(OCIE0A);                                                                               // Timer0: Output Compare A Match Interrupt aktivieren
  TCCR1A = 0;                                                                                         // Timer1: wird fr die Datenausgabe an die LED-Matrix vewendet
  TCCR1B = _BV(WGM12) | _BV(CS10) ;                                                                   // Timer1: Mode 4 (CTC), Vorteiler = 1
  TCCR1C = 0;                                                                                         // Timer1: spezielle Funktionen werden nicht genutzt
  TIMSK1 = _BV(OCIE1A);                                                                               // Timer1: Output Compare A Match Interrupt aktivieren
  TCNT1  = 0;                                                                                         // Timer1: Zhler lschen
  TCCR2A = 0;                                                                                         // Timer2: wird fr die Helligkeitssteuerung der LED-Matrix verwendet
  TCCR2B = 0;                                                                                         // Timer2: Mode 0 (Normal), Vorteiler wird spter gesetzt
  TCCR3A = 0;                                                                                         // Timer3: wird fr die Zeitzhlung, die Taster-Abfrage und den DCF-Empfang verwendet
  TCCR3B = _BV(WGM32) | _BV(CS31) | _BV(CS30);                                                        // Timer3: Mode 4 (CTC), Vorteiler 64 (Takt 312500Hz)
  OCR3A  = F_CPU / 64 / 50 - 1;                                                                       // Timer3: Endwert fr 50Hz setzen
  TIMSK3 = _BV(OCIE3A);                                                                               // Timer3: Output Compare A Match Interrupt aktivieren
  ADMUX  = _BV(ADLAR) | _BV(MUX2) | _BV(MUX1) | _BV(MUX0);                                            // ADC: Referenz AREF, Ergebnis linksbndig, Eingang ADC7
  ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADATE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);                 // ADC: aktivieren, starten, Auto Trigger, Vorteiler = 128
  ADCSRB = 0;                                                                                         // ADC: freilaufend
  DIDR0  = _BV(ADC7D);                                                                                // ADC: Digital Input fr PA7 abschalten
  UCSR1A = 0;                                                                                         // USART1: Status-Register zurcksetzen
  UCSR1B = _BV(RXCIE1) | _BV(RXEN1) | _BV(TXEN1);                                                     // USART1: RXD mit Interrupt und TXD aktivieren
  UCSR1C = _BV(UCSZ11) | _BV(UCSZ10);                                                                 // USART1: Asynchron, 8-Bit, keine Paritt
  UBRR1  = F_CPU / 9600 / 16 - 1;                                                                     // USART1: Geschwindigkeit 9600 Baud einstellen

  lang = eep_read_byte(e_lang);                                       // Sprache aus EEPROM lesen
  if (lang > 1) lang = 0;                                             // wenn Bereich berschritten -> Deutsch einstellen
  dcfsyn = eep_read_byte(e_dcfsyn);                                   // DCF77-Aktivierung aus EEPROM lesen
  if (dcfsyn > 1) dcfsyn = 0;                                         // wenn Bereich berschritten -> ausschalten
  dcfinv = eep_read_byte(e_dcfinv);                                   // DCF77-Invertierung aus EEPROM lesen
  if (dcfinv > 1) dcfinv = 0;                                         // wenn Bereich berschritten -> ausschalten
  dcfpup = eep_read_byte(e_dcfpup);                                   // DCF77-Pull-up aus EEPROM lesen
  if (dcfpup > 1) dcfpup = 0;                                         // wenn Bereich berschritten -> ausschalten
  if (dcfpup) PORTA |= _BV(PINA6);                                    // wenn Pull-up aktiviert -> Pin fr DCF77-Empfnger auf Eingang mit Pull-up setzen
  tstled = eep_read_byte(e_tstled);                                   // Test-LED aus EEPROM lesen
  if (tstled > 2) tstled = 0;                                         // wenn Bereich berschritten -> ausschalten
  colclk = eep_read_byte(e_colclk);                                   // Farbe fr Uhrzeit aus EEPROM lesen
  if (colclk > 16) colclk = 0;                                        // wenn Bereich berschritten -> auf Zufall einstellen
  coldat = eep_read_byte(e_coldat);                                   // Farbe fr Datum, Wochentag und Sensordaten aus EEPROM lesen
  if (coldat > 16) coldat = 0;                                        // wenn Bereich berschritten -> auf Zufall einstellen
  colalm = eep_read_byte(e_colalm);                                   // Farbe fr Alarmmeldungen aus EEPROM lesen
  if (colalm > 16) colalm = 0;                                        // wenn Bereich berschritten -> auf Zufall einstellen
  brimin = eep_read_byte(e_brimin);                                   // Minimale Helligkeit aus EEPROM lesen
  if (brimin > 9) brimin = 0;                                         // wenn Bereich berschritten -> auf 0 setzen
  brimax = eep_read_byte(e_brimax);                                   // Maximale Helligkeit aus EEPROM lesen
  if (brimax > 9) brimax = 9;                                         // wenn Bereich berschritten -> auf 9 setzen
  senfil = eep_read_byte(e_senfil);                                   // Sensorfilter aus EEPROM lesen
  if (senfil > 1) senfil = 0;                                         // wenn Bereich berschritten -> auf 0 setzen
  alarms = eep_read_byte(e_alarms);                                   // Alarmauswertung aus EEPROM lesen
  if (alarms > 1) alarms = 0;                                         // wenn Bereich berschritten -> ausschalten
  for (i = 0; i < 8; i ++) {                                          // 8 Sensornummern bearbeiten
    sennum[i] = eep_read_byte(e_sennum + i);                          // Sensornummer aus EEPROM lesen
    if (sennum[i] > 31) sennum[i] = 0;                                // wenn Bereich berschritten -> auf 0 setzen
  }
  for (i = 0; i < 8; i ++) {                                          // 8 Sensornamen bearbeiten
    for (j = 0; j < 5; j ++) {                                        // jeweils 5 Zeichen bearbeiten
      sennam[i][j] = eep_read_byte(e_sennam + i * 5 + j);             // Zeichen aus EEPROM lesen
      if (sennam[i][j] > 134) sennam[i][j] = 0;                       // wenn Bereich berschritten -> Nullzeichen setzen
    }
    if (sennam[i][0] == 0) {                                          // wenn Sensorname leer
      copy_str(34);                                                   // mit "Sen-" und
      strbuf[4] = i + '1';                                            // Nummer 1-8 vorbelegen
      for (j = 0; j < 5; j ++) sennam[i][j] = strbuf[j];              // auf entsprechenden Platz kopieren
    }
  }
  for (i = 0; i < 4; i ++) {                                          // 4 Alarmzustnde bearbeiten
    almnor[i] = eep_read_byte(e_almnor + i);                          // Alarmzustand aus EEPROM lesen
    if (almnor[i] > 1) almnor[i] = 0;                                 // wenn Bereich berschritten -> auf 0 setzen
  }
  for (i = 0; i < 4; i ++) {                                          // 4 Alarmnamen bearbeiten
    for (j = 0; j < 12; j ++) {                                       // jeweils 12 Zeichen bearbeiten
      almnam[i][j] = eep_read_byte(e_almnam + i * 12 + j);            // Zeichen aus EEPROM lesen
      if (almnam[i][j] > 134) almnam[i][j] = 0;                       // wenn Bereich berschritten -> Nullzeichen setzen
    }
    j = 0;                                                            // Zeichenzhler auf String-Anfang
    if (almnam[i][0] == 0) {                                          // wenn Alarmname leer
      copy_str(35);                                                   // mit "Alarm-" vorbelegen
      do {} while(strbuf[j ++] > 0);                                  // String-Ende suchen
      j--;                                                            // Position korrigieren
      strbuf[j ++] = i + '1';                                         // Alarmnummer ergnzen
      strbuf[j] = 0;                                                  // Nullzeichen ergnzen
      for (j = 0; j < 12; j ++) almnam[i][j] = strbuf[j];             // auf entsprechenden Platz kopieren
    }
  }
  housnd = eep_read_byte(e_housnd);                                   // Stundengong aus EEPROM lesen
  if (housnd > 1) housnd = 1;                                         // wenn Bereich berschritten -> einschalten
  almsnd = eep_read_byte(e_almsnd);                                   // Alarmsound aus EEPROM lesen
  if (almsnd > 1) almsnd = 1;                                         // wenn Bereich berschritten -> einschalten
  sounfr = eep_read_byte(e_sounfr);                                   // Sound von aus dem EEPROM lesen
  if (sounfr > 23) sounfr = 8;                                        // wenn Bereich berschritten -> auf 8 voreinstellen
  sounto = eep_read_byte(e_sounto);                                   // Sound bis aus dem EEPROM lesen
  if (sounto > 23) sounto = 22;                                       // wenn Bereich berschritten -> auf 22 voreinstellen
  msgscr = eep_read_byte(e_msgscr);                                   // Roll-Geschwindigkeit aus dem EEPROM lesen
  if (msgscr > 20) msgscr = 0;                                        // wenn Bereich berschritten -> auf 0 voreinstellen
  avocol = eep_read_byte(e_avocol);                                   // Farbe meiden aus dem EEPROM lesen
  if (avocol > 16) avocol = 0;                                        // wenn Bereich berschritten -> auf 0 voreinstellen (aus)
  shdate = eep_read_byte(e_shdate);                                   // Datum anzeigen aus dem EEPROM lesen
  if (shdate > 1) shdate = 1;                                         // wenn Bereich berschritten -> einschalten
  shwday = eep_read_byte(e_shwday);                                   // Wochentag anzeigen aus dem EEPROM lesen
  if (shwday > 1) shwday = 1;                                         // wenn Bereich berschritten -> einschalten
  chtime = eep_read_byte(e_chtime);                                   // Wechselzeit aus dem EEPROM lesen
  if ((chtime < 2) || (chtime > 4)) chtime = 4;                       // wenn Bereich ber- oder unterschritten -> auf 4s voreinstellen
  autost = eep_read_byte(e_autost);                                   // Automatische Sommerzeit aus dem EEPROM lesen
  if (autost > 1) autost = 1;                                         // wenn Bereich berschritten -> einschalten
  daymod = eep_read_byte(e_daymod);                                   // Taganzeige-Modus aus dem EEPROM lesen
  if (daymod > 2) daymod = 0;                                         // wenn Bereich berschritten -> Kalendertag voreinstellen
  dcfhou = eep_read_byte(e_dcfhou);                                   // DCF-Synchronisierungsstunde aus dem EEPROM lesen
  if (dcfhou > 23) dcfhou = 3;                                        // wenn Bereich berschritten -> auf 3:00 Uhr voreinstellen
  nighfr = eep_read_byte(e_nighfr);                                   // Nacht von aus dem EEPROM lesen
  if (nighfr > 23) nighfr = 0;                                        // wenn Bereich berschritten -> auf 0:00 Uhr voreinstellen
  nighto = eep_read_byte(e_nighto);                                   // Nacht bis aus dem EEPROM lesen
  if (nighto > 23) nighto = 0;                                        // wenn Bereich berschritten -> auf 0:00 Uhr voreinstellen
  if (nighfr != nighto) dcfsyn = 0;                                   // wenn Nachtzeit-Bereich definiert -> feste DCF77-Synchronisierungzeit ausschalten

  if (PINA & _BV(PINA3)) {                                            // wenn Anschluss D der LED-Matrix auf High (ist bei Scan-Modus 1/16 ein Eingang)
    matscn = 0;                                                       // Matrix-Scan-Modus auf 1/16 setzen
    PORTA &= ~_BV(PINA3);                                             // Pull-Up an D ausschalten
    DDRA |= _BV(PINA3);                                               // Port auf Ausgang setzen
    OCR1A  = F_CPU / (200 * 16 * 2) - 1;                              // Timer1: Endwert fr 6400Hz setzen (16-fach Multiplex, 2 Durchlufe, Bildfrequenz 200Hz)
  }
  else {                                                              // wenn Anschluss D der LED-Matrix auf Low (liegt bei Scan-Modus 1/8 fest auf GND = Low)
    matscn = 1;                                                       // Matrix-Scan-Modus auf 1/8 setzen
    OCR1A  = F_CPU / (166 * 16 * 2) - 1;                              // Timer1: Endwert fr 5312Hz setzen (8-fach Multiplex, 2 Durchlufe, Bildfrequenz 322Hz)
  }

  day      = 1;                                                       // Tag zunchst auf 1 setzen
  month    = 1;                                                       // Monat zunchst auf 1 setzen
  year     = c_year;                                                  // Jahr zunchst auf Voreinstellung setzen
  rxwrite  = 0;                                                       // Schreib-Zeiger fr USART-Empfangspuffer initialisieren
  rxread   = 0;                                                       // Lese-Zeiger fr USART-Empfangspuffer initialisieren
  rxpos    = 0;                                                       // USART-Empfangsdaten-Position
  _delay_ms(10);                                                      // 10ms warten, bis ADC stabil arbeitet
  srand(ADC);                                                         // Zufallsgenerator mit ADC-Wert vom Helligkeitssensor initialisieren
  color_random();                                                     // neue Zufallsfarben erzeugen
  disptout = c_start;                                                 // Anzeige-Timeout-Zhler auf Start-Zeit setzen
  dispmode = 3;                                                       // Anzeige-Modus auf Systemstart setzen
  datacoun = chtime - 7;                                              // Datenfeldzhler abhngig von der Wechselzeit voreinstellen, damit nach Ablauf des Systemstarts die Anzeige mit dem Datum beginnt
  bripwm   = 0;                                                       // PWM fr Helligkeit zunchst auf niedrigsten Wert setzen
  sndtout  = 255;                                                     // Soundausgabe inaktiv setzen
  nacttout = 255;                                                     // Nachtaktivierungs-Zhler stoppen
  sndena   = c_sndena;                                                // Soundmodul-Timeout-Zhler setzen
  quitcnt  = 32;                                                      // Quittierungszhler inaktiv setzen
  matrix_fm6126a_init();                                              // Matrix-Initialisierung senden (fr Matrix mit FM6126A erforderlich)
  sei();                                                              // Interrupts freigeben

// --- Beginn Hauptprogrammschleife ---

  while(1) {                                                          // Endlosschleife

    int x;                                                            // universelle Variable
    char hourh, hourl, minuteh, minutel;                              // Zehner/Einer fr Stunden und Minuten
    char secondh, secondl, dayh, dayl;                                // Zehner/Einer fr Sekunden und Kalendertag

    hourh   = hour / 10;                                              // Stunden-Zehner ermitteln
    hourl   = hour % 10;                                              // Stunden-Einer ermitteln
    minuteh = minute / 10;                                            // Minuten-Zehner ermitteln
    minutel = minute % 10;                                            // Minuten-Einer ermitteln
    secondh = second / 10;                                            // Sekunden-Zehner ermitteln
    secondl = second % 10;                                            // Sekunden-Einer ermitteln
    dayh    = day / 10;                                               // Kalendertag-Zehner ermitteln
    dayl    = day % 10;                                               // Kalendertag-Einer ermitteln

    color_clock();                                                    // Farbe fr Uhrzeit einstellen
    posx = 53;                                                        // X-Position fr Sekunden
    posy = 0;                                                         // Y-Position fr Sekunden
    led_digit3(secondh);                                              // Sekunden-Zehner ausgeben
    led_digit3(secondl);                                              // Sekunden-Einer ausgeben
    posx = 0;                                                         // X-Position fr Stunden und Minuten
    posy = 0;                                                         // Y-Position fr Stunden und Minuten

    if (dispmode == 0) {                                              // wenn normaler Anzeige-Modus aktiv
      if (hourh > 0) led_digit1(hourh);                               // wenn Zehner > 0 -> groe Stunden-Zehner ausgeben
      posx = 12;                                                      // X-Position fr Stunden-Einer
      led_digit1(hourl);                                              // groe Stunden-Einer ausgeben
      posx += 4;                                                      // Platz fr Doppelpunkt freihalten
      led_digit1(minuteh);                                            // groe Minuten-Zehner ausgeben
      led_digit1(minutel);                                            // groe Minuten-Einer ausgeben
      if (syncstat == 0) color_alarm();                               // wenn Uhr noch nicht synchronisiert -> Doppelpunkt hat Alarm-Farbe
      if (syncstat == 2) color_date();                                // wenn Sync-Ausfall -> Doppelpunkt hat Datum-Farbe
      if (intcnt < 25) {                                              // wenn erste Sekundenhlfte
        matrix_pixel(24, 5);                                          // groen Doppelpunkt ausgeben (8 Pixel)
        matrix_pixel(24, 6);
        matrix_pixel(25, 5);
        matrix_pixel(25, 6);
        matrix_pixel(24, 11);
        matrix_pixel(24, 12);
        matrix_pixel(25, 11);
        matrix_pixel(25, 12);
      }
      color_date();                                                   // Farbe fr Datum einstellen
      posy = 9;                                                       // Y-Position fr Kalender- oder Wochentag
      if (syncstat) {                                                 // wenn Uhr synchronisiert
        if ((daymod == 0) || ((daymod == 2) && (daymcoun == 0))) {    // wenn Kalendertag-Anzeige gewhlt oder Wechselanzeige und Zhler = 0
          posx = 53;                                                  // X-Position fr Kalendertag
          if (dayh == 0) posx = 56;                                   // wenn Tag einstellig -> X-Position auf Mitte
          else led_chrr(dayh + '0');                                  // sonst Tag-Zehner ausgeben
          led_chrr(dayl + '0');                                       // Tag-Einer ausgeben
        }
        else {                                                        // wenn Wochentag-Anzeige gewhlt oder Wechselanzeige und Zhler = 1
          posx = 58 - strwr2(wday) / 2;                               // X-Position fr zentrierten Wochentag (2 Zeichen)
          led_strr2(wday);                                            // Wochentag ausgeben (2 Zeichen)
        }
      }
      else {                                                          // wenn Uhr noch nicht synchronisiert
        posx = 56;                                                    // X-Position auf Mitte
        led_chrr('?');                                                // Fragezeichen ausgeben
      }
    }
    else {                                                            // bei allen anderen Anzeigemodi
      if (hourh > 0) led_digit2(hourh);                               // wenn Stunden-Zehner > 0 -> kleine Stunden-Zehner ausgeben
      posx = 12;                                                      // X-Position fr Stunden-Einer
      led_digit2(hourl);                                              // kleine Stunden-Einer ausgeben
      posx += 4;                                                      // Platz fr Doppelpunkt freihalten
      led_digit2(minuteh);                                            // kleine Minuten-Zehner ausgeben
      led_digit2(minutel);                                            // kleine Minuten-Einer ausgeben
      if (syncstat == 0) color_alarm();                               // wenn Uhr noch nicht synchronisiert -> Doppelpunkt hat Alarm-Farbe
      if (syncstat == 2) color_date();                                // wenn Sync-Ausfall -> Doppelpunkt hat Datum-Farbe
      if (intcnt < 25) {                                              // wenn erste Sekundenhlfte
        matrix_pixel(24, 1);                                          // kleinen Doppelpunkt ausgeben (8 Pixel)
        matrix_pixel(24, 2);
        matrix_pixel(25, 1);
        matrix_pixel(25, 2);
        matrix_pixel(24, 5);
        matrix_pixel(24, 6);
        matrix_pixel(25, 5);
        matrix_pixel(25, 6);
      }
    }

    if (dispmode < 3) {                                               // wenn normale, Alarm- oder Nachrichten-Anzeige aktiv
      color_date();                                                   // Farbe fr Datum einstellen
      posy = 21;                                                      // Y-Position fr Datenfeld bei normaler Anzeige
      if (dispmode > 0) posy = 10;                                    // Y-Position fr Datenfeld bei Alarm- und Nachrichten-Anzeige
      if (datacoun == 0)  {                                           // wenn Datenfeld 0 ausgewhlt (Datum)
        if (syncstat == 0) {                                          // wenn noch keine Synchronisierung
          posx = 32 - strw(9) / 2;                                    // X-Position fr zentrierten Text "kein Datum"
          led_str(9);                                                 // Text ausgeben
        }
        else {                                                        // wenn Uhr synchronisiert
          posx = 32 - (chrw('0') * 8 + chrw('.') * 2 + 9) / 2;        // X-Position fr zentriertes Datum (8 Ziffern, 2 Punkte und 9 Zwischenrume)
          led_chr(dayh + '0');                                        // Tag-Zehner ausgeben
          led_chr(dayl + '0');                                        // Tag-Einer ausgeben
          led_chr('.');                                               // Punkt ausgeben
          led_chr(month / 10 + '0');                                  // Monat-Zehner ausgeben
          led_chr(month % 10 + '0');                                  // Monat-Einer ausgeben
          led_chr('.');                                               // Punkt ausgeben
          led_chr(c_yearh / 10 + '0');                                // Jahrhundert-Zehner ausgeben
          led_chr(c_yearh % 10 + '0');                                // Jahrhundert-Einer ausgeben
          led_chr(year / 10 + '0');                                   // Jahr-Zehner ausgeben
          led_chr(year % 10 + '0');                                   // Jahr-Einer ausgeben
        }
      }
      if (datacoun == 1) {                                            // wenn Datenfeld 1 ausgewhlt (Wochentag)
        if (syncstat == 0) {                                          // wenn noch keine Synchronisierung
          posx = 32 - strw(10) / 2;                                   // X-Position fr zentrierten Text "kein Tag"
          led_str(10);                                                // Text ausgeben
        }
        else {                                                        // wenn Uhr synchronisiert
          posx = 32 - strw(wday) / 2;                                 // X-Position fr zentrierten Wochentag
          led_str(wday);                                              // Wochentag ausgeben
        }
      }
      if ((datacoun > 1) && (datacoun < 33)) {                        // wenn Datenfeld zwischen 2 und 32 ausgewhlt (Sensorwert oder Geburtstag)
        if (senfil) {                                                 // wenn Sensorfilter eingeschaltet
          if (datacoun < 10) {                                        // wenn Zhler im gltigen Bereich (2-9)
            posx = 32 - senw(sennum[datacoun - 2] - 1) / 2;           // korrigierten Zhler ber Filter als Index fr das Sensorfeld verwenden, X-Position fr zentrierten Text ermitteln
            led_sen(sennum[datacoun - 2] - 1);                        // Sensorwert ausgeben
          }
        }
        else {                                                        // wenn Sensorfilter ausgeschaltet
          posx = 32 - senw(datacoun - 2) / 2;                         // korrigierten Zhler als Index fr das Sensorfeld verwenden, X-Position fr zentrierten Text ermitteln
          led_sen(datacoun - 2);                                      // Sensorwert oder Geburtstag ausgeben
        }
      }
      if (datacoun == 33) {                                           // wenn Datenfeld 33 ausgewhlt (keine Daten)
        posx = 32 - strw(52) / 2;                                     // X-Position fr zentrierten String "keine Daten"
        led_str(52);                                                  // Text ausgeben
      }
      posy = 19;                                                      // Y-Position fr Quittierungsanimation oder Stummschaltungszeile bei normaler Anzeige
      if (dispmode > 0) posy = 9;                                     // Y-Position fr Quittierungsanimation oder Stummschaltungszeile bei Alarm- und Nachrichten-Anzeige
      color_alarm();                                                  // Farbe fr Alarm einstellen
      if (quitcnt < 32) {                                             // wenn Quittierungszhler aktiv
        for (i = quitcnt; i < 32; i ++) {                             // Schleife fr 2 Pixelhalbreihen
          matrix_pixel(i, posy);                                      // Pixel fr linke Reihe ausgeben
          matrix_pixel(63 - i, posy);                                 // Pixel fr rechte Reihe ausgeben
        }
      }
      else {                                                          // wenn Qittierungszhler nicht aktiv
        if (mute) {                                                   // wenn Stummschaltung aktiv
          for (i = 0; i < 64; i ++) matrix_pixel(i, posy);            // Pixelzeile ausgeben
        }
      }
    }

    if (dispmode == 1) {                                              // wenn Alarm-Modus aktiv
      color_alarm();                                                  // Farbe fr Alarme einstellen
      posx = 32 - almw(almcoun) / 2;                                  // X-Position fr Alarmfeld (zentriert)
      posy = 21;                                                      // Y-Position fr Alarmfeld
      if (almstat[almcoun] == 0) undrline = 1;                        // wenn Alarm nicht mehr aktiv -> Unterstreichung einschalten
      led_alm(almcoun);                                               // Alarmfeld ausgeben
      undrline = 0;                                                   // Unterstreichung wieder ausschalten
    }

    uint8_t s1, s2;                                                   // Zwischenspeicher fr Spaltenbytes
    if (dispmode == 2) {                                              // wenn Nachrichten-Modus aktiv
      color_alarm();                                                  // Farbe fr Alarme einstellen
      for (i = 0; i < 64; i ++) {                                     // 64 Pixelspalten ausgeben
        if ((msgpbpos + (i * 2) < 0) ||
          (msgpbpos + (i * 2) + 2 > msgpbend)) {                      // wenn aktuelle Pixelspalte vor oder nach der Nachricht
          s1 = 0;                                                     // Spaltenbyte 1 lschen
          s2 = 0;                                                     // Spaltenbyte 2 lschen
        }
        else {                                                        // wenn aktuelle Pixelspalte innerhalb der Nachricht
          s1 = msgpbuf[msgpbpos + (i * 2)];                           // Spaltenbyte 1 lesen
          s2 = msgpbuf[msgpbpos + (i * 2) + 1];                       // Spaltenbyte 2 lesen
        }
        for (j = 0; j < 3; j ++) {                                    // Spaltenbyte 1, Pixelzeile 0-2 bearbeiten
          if (s1 & 1) matrix_pixel(i, 21 + 2 - j);                    // wenn Pixel gesetzt -> in LED-Matrix setzen
          s1 >>= 1;                                                   // Spaltenbyte auf nchste Bit-Position schieben
        }
        for (j = 0; j < 8; j ++) {                                    // Spaltenbyte 2, Pixelzeile 0-7 (entspricht 3-10) bearbeiten
          if (s2 & 1) matrix_pixel(i, 21 + 10 - j);                   // wenn Pixel gesetzt -> in LED-Matrix setzen
          s2 >>= 1;                                                   // Spaltenbyte auf nchste Bit-Position schieben
        }
      }
    }

    if (dispmode == 3) {                                              // wenn Systemstart-Modus aktiv
      color_date();                                                   // Farbe fr Datum einstellen
      posx = 32 - strw(8) / 2;                                        // X-Position fr Zeile 1 der Versionsinformationen (zentrierter Text "Matrix-Uhr")
      posy = 10;                                                      // Y-Position fr Zeile 1 der Versionsinformationen
      led_str(8);                                                     // "Matrix-Uhr" ausgeben
      color_alarm();                                                  // Farbe fr Alarme einstellen
      posx = 32 - strw(0) / 2;                                        // X-Position fr Zeile 2 der Versionsinformationen (zentrierte Versionsnummer)
      posy = 21;                                                      // Y-Position fr Zeile 2 der Versionsinformationen
      led_str(0);                                                     // Versionsnummer ausgeben
    }

    if (dispmode == 4) {                                              // wenn Einstellungs-Modus aktiv
      color_date();                                                   // Farbe fr Datum einstellen (die Einstellung in der mittleren Zeile wird immer in Datum-Farbe dargestellt)
      posx = 0;                                                       // X-Position fr Einstellung
      posy = 10;                                                      // Y-Position fr Einstellung
      switch (setmode) {                                              // weiterer Verlauf ist von der Einstellung abhngig
        case 0:                                                       // wenn Einstellung 0 (Farbe fr Uhrzeit)
          led_str(12);                                                // Text "Zeit-Farbe" ausgeben
          posy = 21;                                                  // Y-Position fr Parameter
          posx = 32 - (chrw(137) * 9 + 8) / 2;                        // X-Position fr Parameter (Farbreihe mit 9 Blcken und Zwischenraum) zentriert
          for (i = colclk - 4; i < colclk + 5; i ++) {                // 4 vorherige Farben, aktuelle Farbe und 4 nachfolgende Farben
            switch (i) {                                              // weitere Anzeige abhngig vom Zhlerwert
              case -4 ... -1:                                         // falls Zhler negativ (entspricht Farbwert 12-15)
                matrix_color(i + 16);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              case 0:                                                 // falls Zhler = 0 und
              case 17:                                                // falls Zhler = 17 (berlauf, gleicher Ablauf wie 0)
                matrix_color(colclkr);                                // Zufallsfarbe einstellen
                led_chr(136);                                         // leeren Pixelblock ausgeben
                break;
              case 18 ... 20:                                         // falls Zhler > 17 (berlauf, entspricht Farbwert 0-2)
                matrix_color(i - 18);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              default:                                                // bei allen anderen Zhlerwerten
                matrix_color(i - 1);                                  // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
            }
          }
          color_alarm();                                              // Farbe fr Alarme einstellen (der Parameter in der unteren Zeile wird immer in Alarm-Farbe dargestellt)
          posx = 32 - chrw(135) / 2;                                  // X-Position fr Farbzeiger zentriert
          led_chr(135);                                               // Farbzeiger ausgeben
          break;
        case 1:                                                       // wenn Einstellung 1 (Farbe fr Datum)
          led_str(13);                                                // Text "Datum-Farbe" ausgeben
          posy = 21;                                                  // Y-Position fr Parameter
          posx = 32 - (chrw(137) * 9 + 8) / 2;                        // X-Position fr Parameter (Farbreihe mit 9 Blcken und Zwischenraum) zentriert
          for (i = coldat - 4; i < coldat + 5; i ++) {                // 4 vorherige Farben, aktuelle Farbe und 4 nachfolgenden Farben
            switch (i) {                                              // weitere Anzeige abhngig vom Zhlerwert
              case -4 ... -1:                                         // falls Zhler negativ (entspricht Farbwert 12-15)
                matrix_color(i + 16);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              case 0:                                                 // falls Zhler = 0 und
              case 17:                                                // falls Zhler = 17 (berlauf, gleicher Ablauf wie 0)
                matrix_color(coldatr);                                // Zufallsfarbe einstellen
                led_chr(136);                                         // leeren Pixelblock ausgeben
                break;
              case 18 ... 20:                                         // falls Zhler > 17 (berlauf, entspricht Farbwert 0-2)
                matrix_color(i - 18);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              default:                                                // bei allen anderen Zhlerwerten
                matrix_color(i - 1);                                  // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
            }
          }
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 32 - chrw(135) / 2;                                  // X-Position fr Farbzeiger zentriert
          led_chr(135);                                               // Farbzeiger ausgeben
          break;
        case 2:                                                       // wenn Einstellung 2 (Farbe fr Alarme)
          led_str(14);                                                // Text "Alarm-Farbe" ausgeben
          posy = 21;                                                  // Y-Position fr Parameter
          posx = 32 - (chrw(137) * 9 + 8) / 2;                        // X-Position fr Parameter (Farbreihe mit 9 Blcken und Zwischenraum) zentriert
          for (i = colalm - 4; i < colalm + 5; i ++) {                // 4 vorherige Farben, aktuelle Farbe und 4 nachfolgenden Farben
            switch (i) {                                              // weitere Anzeige abhngig vom Zhlerwert
              case -4 ... -1:                                         // falls Zhler negativ (entspricht Farbwert 12-15)
                matrix_color(i + 16);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              case 0:                                                 // falls Zhler = 0 und
              case 17:                                                // falls Zhler = 17 (berlauf, gleicher Ablauf wie 0)
                matrix_color(colalmr);                                // Zufallsfarbe einstellen
                led_chr(136);                                         // leeren Pixelblock ausgeben
                break;
              case 18 ... 20:                                         // falls Zhler > 17 (berlauf, entspricht Farbwert 0-2)
                matrix_color(i - 18);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              default:                                                // bei allen anderen Zhlerwerten
                matrix_color(i - 1);                                  // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
            }
          }
          color_date();                                               // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 32 - chrw(135) / 2;                                  // X-Position fr Farbzeiger zentriert
          led_chr(135);                                               // Farbzeiger ausgeben
          break;
        case 3:                                                       // wenn Einstellung 3 (Farbe meiden)
          led_str(48);                                                // Text "Farbe meiden" ausgeben
          posy = 21;                                                  // Y-Position fr Parameter
          posx = 32 - (chrw(137) * 9 + 8) / 2;                        // X-Position fr Parameter (Farbreihe mit 9 Blcken und Zwischenraum) zentriert
          for (i = avocol - 4; i < avocol + 5; i ++) {                // 4 vorherige Farben, aktuelle Farbe und 4 nachfolgenden Farben
            switch (i) {                                              // weitere Anzeige abhngig vom Zhlerwert
              case -4 ... -1:                                         // falls Zhler negativ (entspricht Farbwert 12-15)
                matrix_color(i + 16);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              case 0:                                                 // falls Zhler = 0 und
              case 17:                                                // falls Zhler = 17 (berlauf, gleicher Ablauf wie 0)
                matrix_color(15);                                     // Wei einstellen
                led_chr(136);                                         // leeren Pixelblock ausgeben
                break;
              case 18 ... 20:                                         // falls Zhler > 17 (berlauf, entspricht Farbwert 0-2)
                matrix_color(i - 18);                                 // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
                break;
              default:                                                // bei allen anderen Zhlerwerten
                matrix_color(i - 1);                                  // Farbwert korrigieren und einstellen
                led_chr(137);                                         // vollen Pixelblock ausgeben
            }
          }
          color_date();                                               // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 32 - chrw(135) / 2;                                  // X-Position fr Farbzeiger zentriert
          led_chr(135);                                               // Farbzeiger ausgeben
          break;
        case 4:                                                       // wenn Einstellung 4 (Minimale Helligkeit)
          led_str(27);                                                // Text "Helligkeit Min" ausgeben
          posy = 21;                                                  // Y-Position fr Helligkeitswert
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstelligen Helligkeitswert (rechtbndig)
          if (brimin > 8) {                                           // wenn zweistellig (ab Wert 9)
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistelligen Helligkeitswert (rechtsbndig)
            led_chr((brimin + 1) / 10 + '0');                         // Helligkeit-Zehner ausgeben (statt 0-9 -> 1-10)
          }
          led_chr((brimin + 1) % 10 + '0');                           // Helligkeit-Einer ausgeben (statt 0-9 -> 1-10)
          break;
        case 5:                                                       // wenn Einstellung 5 (Maximale Helligkeit)
          led_str(28);                                                // Text "Helligkeit Max" ausgeben
          posy = 21;                                                  // Y-Position fr Helligkeitswert
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstelligen Helligkeitswert (rechtsbndig)
          if (brimax > 8) {                                           // wenn zweistellig (ab Wert 9)
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistelligen Helligkeitswert (rechtsbndig)
            led_chr((brimax + 1) / 10 + '0');                         // Helligkeit-Zehner ausgeben (statt 0-9 -> 1-10)
          }
          led_chr((brimax + 1) % 10 + '0');                           // Helligkeit-Einer ausgeben (statt 0-9 -> 1-10)
          break;
        case 6:                                                       // wenn Einstellung 6 (Test-LED)
          led_str(29);                                                // Text "Test-LED" ausgeben
          if (tstled == 0) posx = 64 - strw(25);                      // wenn Parameter "aus" -> X-Position fr Text "aus" (rechtsbndig)
          else posx = 64 - strw(tstled + 41);                         // sonst X-Position fr Text "DCF/RS-232" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          if (tstled == 0) led_str(25);                               // wenn Parameter "aus" -> ausgeben
          else led_str(tstled + 41);                                  // sonst Parameter "DCF/RS-232" ausgeben
          break;
        case 7:                                                       // wenn Einstellung 7 (Feste DCF-Synchronisierungszeit)
          led_str(16);                                                // Text "Feste Sync-Z" ausgeben
          posx = 64 - strw(dcfsyn + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(dcfsyn + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 8:                                                       // wenn Einstellung 8 (DCF-Invertierung)
          led_str(17);                                                // Text "DCF-Invert." ausgeben
          posx = 64 - strw(dcfinv + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(dcfinv + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 9:                                                       // wenn Einstellung 9 (DCF-Pull-up)
          led_str(18);                                                // Text "DCF-Pull-up" ausgeben
          posx = 64 - strw(dcfpup + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(dcfpup + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 10:                                                      // wenn Einstellung 10 (Stundengong)
          led_str(36);                                                // Text "Stundengong" ausgeben
          posx = 64 - strw(housnd + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(housnd + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 11:                                                      // wenn Einstellung 11 (Alarmsound)
          led_str(37);                                                // Text "Alarmsound" ausgeben
          posx = 64 - strw(almsnd + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(almsnd + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 12:                                                      // wenn Einstellung 12 (Sound von)
          led_str(38);                                                // Text "Sound von" ausgeben
          posy = 21;                                                  // Y-Position fr Stunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (sounfr > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(sounfr / 10 + '0');                               // Stunden-Zehner ausgeben
          }
          led_chr(sounfr % 10 + '0');                                 // Stunden-Einer ausgeben
          break;
        case 13:                                                      // wenn Einstellung 13 (Sound bis)
          led_str(39);                                                // Text "Sound bis" ausgeben
          posy = 21;                                                  // Y-Position fr Stunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (sounto > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(sounto / 10 + '0');                               // Stunden-Zehner ausgeben
          }
          led_chr(sounto % 10 + '0');                                 // Stunden-Einer ausgeben
          break;
        case 14:                                                      // wenn Einstellung 14 (Roll-Geschwindigkeit)
          led_str(40);                                                // Text "Roll-Geschw." ausgeben
          posy = 21;                                                  // Y-Position fr Rollwert
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstelligen Rollwert (rechtsbndig)
          if (msgscr > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistelligen Rollwert (rechtsbndig)
            led_chr(msgscr / 10 + '0');                               // Rollwert-Zehner ausgeben
          }
          led_chr(msgscr % 10 + '0');                                 // Rollwert-Einer ausgeben
          for (i = 0; i < 8; i ++) {                                  // Testmuster im Abstand von 8 Pixel ausgeben
            if ((msgstest & 7) == i) {                                // untere 3 Bit vom Roll-Testzhler filtern
              matrix_pixel(i, 29);                                    // und an entsprechender Position insgesamt 6 Pixel ausgeben
              matrix_pixel(i + 8, 29);
              matrix_pixel(i + 16, 29);
              matrix_pixel(i + 24, 29);
              matrix_pixel(i + 32, 29);
              matrix_pixel(i + 40, 29);
            }
          }
          break;
        case 15:                                                      // wenn Einstellung 15 (Sensorfilter)
          led_str(32);                                                // Text "Sensorfilter" ausgeben
          posx = 64 - strw(senfil + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr aus/ein
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(senfil + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 16 ... 23:                                               // wenn Einstellung 16-23 (Sensornummer 1-8)
          led_str(33);                                                // Text "Sensornum. " ausgeben
          led_chr(setmode - 15 + '0');                                // Platznummer ausgeben (1-8)
          posy = 21;                                                  // Y-Position fr Sensornummer
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          if (sennum[setmode - 16] == 0) {                            // wenn Sensornummer = 0 (aus)
            posx = 64 - strw(25);                                     // X-Position fr Text "aus" (rechtsbndig)
            led_str(25);                                              // Text "aus" ausgeben
          }
          else {                                                      // wenn Sensornummer > 0
            posx = 64 - chrw('0');                                    // X-Position fr einstellige Nummer (rechtsbndig)
            if (sennum[setmode - 16] > 9) {                           // wenn zweistellig
              posx = 64 - chrw('0') * 2 - 1;                          // X-Position fr zweistellige Nummer (rechtsbndig)
              led_chr(sennum[setmode - 16] / 10 + '0');               // Sensornummer-Zehner ausgeben
            }
            led_chr(sennum[setmode - 16] % 10 + '0');                 // Sensornummer-Einer ausgeben
          }
          break;
        case 24 ... 31:                                               // wenn Einstellung 24-31 (Sensorname 1-8)
          led_str(47);                                                // Text "Sensornam. " ausgeben
          led_chr(setmode - 23 + '0');                                // Platznummer ausgeben (1-8)
          posx = 64;                                                  // X-Position auf Zeilenende
          posy = 21;                                                  // Y-Position fr Sensorname
          x = 0;                                                      // Zeichenzhler lschen
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          for (i = 0; i < 5; i ++) {                                  // 5 Zeichen bearbeiten (Textlnge ermitteln)
            x = x + chrw(sennam[setmode - 24][i]) + 1;                // Zeichenlnge und Zwischenraum addieren
          }
          posx = 64 - x;                                              // Textlnge von Endpositon abziehen (rechtsbndig)
          for (i = 0; i < 5; i ++) {                                  // 5 Zeichen bearbeiten (Zeichen ausgeben)
            if (editmode && (editpos == i)) undrline = 1;             // wenn Editier-Modus und gerade editierte Zeichenposition -> Unterstreichung einschalten
            led_chr(sennam[setmode - 24][i]);                         // Zeichen ausgeben
            undrline = 0;                                             // Unterstreichung wieder ausschalten
          }
          break;
        case 32:                                                      // wenn Einstellung 32 (Alarme)
          led_str(30);                                                // Text "Alarme" ausgeben
          posx = 64 - strw(alarms + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(alarms + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 33 ... 36:                                               // wenn Einstellung 33-36 (Alarmname 1-4)
          led_str(31);                                                // Text "Alarmname " ausgeben
          led_chr(setmode - 32 + '0');                                // Platznummer ausgeben (1-4)
          posy = 21;                                                  // Y-Position fr Alarmname
          x = 0;                                                      // Zeichenzhler lschen
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          for (i = 0; i < 12; i ++) {                                 // 12 Zeichen bearbeiten (Textlnge ermitteln)
            x = x + chrw(almnam[setmode - 33][i]) + 1;                // Zeichenlnge und Zwischenraum addieren
          }
          posx = 64 - x;                                              // Textlnge von Endpositon abziehen (rechtsbndig)
          if (x > 63) {                                               // wenn Textlnge ber 63
            if (editpos < 6) posx = 1;                                // wenn die Zeichen 1-6 bearbeitet werden -> auf linksbndig umschalten
          }
          for (i = 0; i < 12; i ++) {                                 // 12 Zeichen bearbeiten (Zeichen ausgeben)
            if (editmode && (editpos == i)) undrline = 1;             // wenn Editier-Modus und gerade editierte Zeichenposition -> Unterstreichung einschalten
            led_chr(almnam[setmode - 33][i]);                         // Zeichen ausgeben
            undrline = 0;                                             // Unterstreichung wieder ausschalten
          }
          break;
        case 37:                                                      // wenn Einstellung 37 (Tag)
          led_str(19);                                                // Text "Tag" ausgeben
          posy = 21;                                                  // Y-Position fr Tag
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstelligen Tag (rechtsbndig)
          if (dayh > 0) {                                             // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistelligen Tag (rechtsbndig)
            led_chr(dayh + '0');                                      // Tag-Zehner ausgeben
          }
          led_chr(dayl + '0');                                        // Tag-Einer ausgeben
          break;
        case 38:                                                      // wenn Einstellung 38 (Monat)
          led_str(20);                                                // Text "Monat" ausgeben
          posy = 21;                                                  // Y-Position fr Monat
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstelligen Monat (rechtsbndig)
          if (month > 9) {                                            // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistelligen Monat (rechtsbndig)
            led_chr(month / 10 + '0');                                // Monat-Zehner ausgeben
          }
          led_chr(month % 10 + '0');                                  // Monat-Einer ausgeben
          break;
        case 39:                                                      // wenn Einstellung 39 (Jahr)
          led_str(21);                                                // Text "Jahr" ausgeben
          posx = 64 - chrw('0') * 4 - 3;                              // X-Position fr vierstelliges Jahr (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Jahr
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_chr(c_yearh / 10 + '0');                                // Jahrhundert-Zehner ausgeben
          led_chr(c_yearh % 10 + '0');                                // Jahrhundert-Einer ausgeben
          led_chr(year / 10 + '0');                                   // Jahr-Zehner ausgeben
          led_chr(year % 10 + '0');                                   // Jahr-Einer ausgeben
          break;
        case 40:                                                      // wenn Einstellung 40 (Stunde)
          led_str(22);                                                // Text "Stunde" ausgeben
          posy = 21;                                                  // Y-Position fr Stunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (hourh > 0) {                                            // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(hourh + '0');                                     // Stunden-Zehner ausgeben
          }
          led_chr(hourl + '0');                                       // Stunden-Einer ausgeben
          break;
        case 41:                                                      // wenn Einstellung 41 (Minute)
          led_str(23);                                                // Text "Minute" ausgeben
          posy = 21;                                                  // Y-Position fr Minute
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Minute (rechtsbndig)
          if (minuteh > 0) {                                          // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Minute (rechtsbndig)
            led_chr(minuteh + '0');                                   // Minuten-Zehner ausgeben
          }
          led_chr(minutel + '0');                                     // Minuten-Einer ausgeben
          break;
        case 42:                                                      // wenn Einstellung 42 (Sekunde)
          led_str(24);                                                // Text "Sekunde" ausgeben
          posy = 21;                                                  // Y-Position fr Sekunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Sekunde (rechtsbndig)
          if (secondh > 0) {                                          // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Sekunde (rechtsbndig)
            led_chr(secondh + '0');                                   // Sekunden-Zehner ausgeben
          }
          led_chr(secondl + '0');                                     // Sekunden-Einer ausgeben
          break;
        case 43:                                                      // wenn Einstellung 43 (LED-Matrix-Test)
          if (matrtest) {                                             // wenn LED-Matrix-Test-Modus aktiv
            matrix_color(16);                                         // Farbe auf "Testwei" einstellen
            for (i = 0; i < 64; i ++) {                               // alle 64 Pixelspalten
              for (j = 0; j < 32; j ++) {                             // alle 32 Pixelzeilen
                matrix_pixel(i, j);                                   // Pixel setzen
              }
            }
          }
          else {                                                      // wenn LED-Matrix-Test-Modus nicht aktiv
            led_str(41);                                              // Text "Matrix-Test" ausgeben
            posx = 64 - strw(46);                                     // X-Position fr Text "Start" (rechtsbndig)
            posy = 21;                                                // Y-Position fr Text "Start"
            color_alarm();                                            // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
            led_str(46);                                              // Parameter "Start" ausgeben
          }
          break;
        case 44:                                                      // wenn Einstellung 44 (Sound-Test)
          led_str(44);                                                // Text "Sound " ausgeben
          posy = 21;                                                  // Y-Position fr Minute
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - strw(45) - chrw('0') - 1;                       // X-Position fr Text "Sound " und eine Ziffer (rechtsbndig)
          led_str(45);                                                // Parameter "Sound " ausgeben
          led_chr(sndtest + '1');                                     // Soundnummer ausgeben
          break;
        case 45:                                                      // wenn Einstellung 45 (Sprache)
          led_str(15);                                                // Text "Sprache" ausgeben
          posx = 64 - strw(11);                                       // X-Position fr Text "Deutsch" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Sprache-Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(11);                                                // Text "Deutsch" ausgeben
          break;
        case 46:                                                      // wenn Einstellung 46 (Datum anzeigen)
          led_str(50);                                                // Text "Datum zeigen" ausgeben
          posx = 64 - strw(shdate + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(shdate + 25);                                       // Parameter "aus/ein" -> ausgeben
          break;
        case 47:                                                      // wenn Einstellung 47 (Wochentag anzeigen)
          led_str(51);                                                // Text "W-Tag zeigen" ausgeben
          posx = 64 - strw(shwday + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(shwday + 25);                                       // Parameter "aus/ein" ausgeben
          break;
        case 48:                                                      // wenn Einstellung 48 (Wechselzeit)
          led_str(53);                                                // Text "Wechselzeit" ausgeben
          posx = 64 - strw(54) - chrw(' ') - chrw('0') - 2;           // X-Position fr Text "n Sekunden" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_chr(chtime + '0');                                      // Wechselzeit ausgeben
          led_chr(' ');                                               // Leerzeichen ausgeben
          led_str(54);                                                // Text "Sekunden" ausgeben
          break;
        case 49:                                                      // wenn Einstellung 49 (Automatische Sommerzeit)
          led_str(55);                                                // Text "Aut. Sommerz." ausgeben
          posx = 64 - strw(autost + 25);                              // X-Position fr Text "aus/ein" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(autost + 25);                                       // Parameter "aus/ein" -> ausgeben
          break;
        case 50:                                                      // wenn Einstellung 50 (Taganzeige)
          led_str(56);                                                // Text "Taganzeige" ausgeben
          posx = 64 - strw(daymod + 57);                              // X-Position fr Text "Kalendertag/Wochentag/Wechsel" (rechtsbndig)
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          led_str(daymod + 57);                                       // Parameter "Kalendertag/Wochentag/Wechsel" ausgeben
          break;
        case 51:                                                      // wenn Einstellung 51 (DCF77-Synchronisierungsstunde)
          led_str(60);                                                // Text "Sync-Stunde" ausgeben
          posy = 21;                                                  // Y-Position fr Parameter
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (dcfhou > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(dcfhou / 10 + '0');                               // Stunden-Zehner ausgeben
          }
          led_chr(dcfhou % 10 + '0');                                 // Stunden-Einer ausgeben
          break;
        case 52:                                                      // wenn Einstellung 52 (Nacht von)
          led_str(61);                                                // Text "Nacht von" ausgeben
          posy = 21;                                                  // Y-Position fr Stunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (nighfr > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(nighfr / 10 + '0');                               // Stunden-Zehner ausgeben
          }
          led_chr(nighfr % 10 + '0');                                 // Stunden-Einer ausgeben
          break;
        case 53:                                                      // wenn Einstellung 53 (Nacht bis)
          led_str(62);                                                // Text "Nacht bis" ausgeben
          posy = 21;                                                  // Y-Position fr Stunde
          color_alarm();                                              // Farbe fr Alarme einstellen (wird fr Parameter verwendet)
          posx = 64 - chrw('0');                                      // X-Position fr einstellige Stunde (rechtsbndig)
          if (nighto > 9) {                                           // wenn zweistellig
            posx = 64 - chrw('0') * 2 - 1;                            // X-Position fr zweistellige Stunde (rechtsbndig)
            led_chr(nighto / 10 + '0');                               // Stunden-Zehner ausgeben
          }
          led_chr(nighto % 10 + '0');                                 // Stunden-Einer ausgeben
          break;
      }
    }
    matrix_swap();                                                    // Matrix-Puffer umschalten
    matrix_clear();                                                   // aktuellen Matrix-Puffer lschen

    uint8_t errcnt;                                                   // Fehlerzhler fr DCF77-Auswertung
    uint8_t parcnt;                                                   // Zhler fr die Parittsprfung der Daten
    uint8_t dcftmp;                                                   // Zwischenspeicher fr den DCF77-Impulszhler
    uint8_t zeroflg;                                                  // Nullbyte-Flag (wenn DCF77-Puffer nur 0-Werte enthlt)

    if (dcfstat2) {                                                   // wenn ein DCF77-Minutenzyklus komplett ist
      cli();                                                          // Interrupts sperren
      dcfstat2 = 0;                                                   // Flag fr Minutenzyklus lschen
      dcftmp = dcfimp;                                                // DCF77-Impulszhler kopieren
      dcfimp = 0;                                                     // DCF77-Impulszhler lschen
      sei();                                                          // Interrupts wieder freigeben

      if (dcftmp == 59) {                                             // wenn 59 Impulse empfangen worden sind
        errcnt = 0;                                                   // Fehlerzhler lschen
        parcnt = 0;                                                   // Parittszhler lschen
        for (i = 4; i < 12; i ++) {                                   // Minuten-Bits (im Byte 04-11) prfen
          if (dcftab[i]) parcnt ++;                                   // wenn Bit = 1 -> Parittszhler erhhen
        }
        if (parcnt & 1) errcnt ++;                                    // wenn Minuten-Prfsumme ungeradzahlig -> Fehler
        parcnt = 0;                                                   // Parittszhler lschen
        for (i = 12; i < 19; i ++) {                                  // Stunden-Bits (im Byte 12-18) prfen
          if (dcftab[i]) parcnt ++;                                   // wenn Bit = 1 -> Parittszhler erhhen
        }
        if (parcnt & 1) errcnt ++;                                    // wenn Stunden-Prfsumme ungeradzahlig -> Fehler
        parcnt = 0;                                                   // Parittszhler lschen
        for (i = 19; i < 42; i ++) {                                  // Datum-Bits (im Byte 19-41) prfen
          if (dcftab[i]) parcnt ++;                                   // wenn Bit = 1 -> Parittszhler erhhen
        }
        if (parcnt & 1) errcnt ++;                                    // wenn Datum-Prfsumme ungeradzahlig -> Fehler

        if (errcnt == 0) {                                            // wenn alle Prfbits ok -> Zeit ermitteln
          dcfbuf1[0] = dcftab[0];                                     // Zeitzone speichern
          dcfbuf1[1] = dcftab[4] +                                    // Minuten-Bits zusammenfgen
            (dcftab[5] * 2) +
            (dcftab[6] * 4) +
            (dcftab[7] * 8) +
            (dcftab[8] * 10) +
            (dcftab[9] * 20) +
            (dcftab[10] * 40);                                        // und Minutenwert speichern

          dcfbuf1[2] = dcftab[12] +                                   // Stunden-Bits zusammenfgen
            (dcftab[13] * 2) +
            (dcftab[14] * 4) +
            (dcftab[15] * 8) +
            (dcftab[16] * 10) +
            (dcftab[17] * 20);                                        // und Stundenwert speichern

          dcfbuf1[3] = dcftab[19] +                                   // Kalendertag-Bits zusammenfgen
            (dcftab[20] * 2) +
            (dcftab[21] * 4) +
            (dcftab[22] * 8) +
            (dcftab[23] * 10) +
            (dcftab[24] * 20);                                        // und Kalendertag speichern

          dcfbuf1[4] = dcftab[25] +                                   // Wochentag-Bits zusammenfgen
            (dcftab[26] * 2) +
            (dcftab[27] * 4);                                         // und Wochentag speichern

          dcfbuf1[5] = dcftab[28] +                                   // Monat-Bits zusammenfgen
            (dcftab[29] * 2) +
            (dcftab[30] * 4) +
            (dcftab[31] * 8) +
            (dcftab[32] * 10);                                        // und Monat speichern

          dcfbuf1[6] = dcftab[33] +                                   // Jahr-Bits zusammenfgen
            (dcftab[34] * 2) +
            (dcftab[35] * 4) +
            (dcftab[36] * 8) +
            (dcftab[37] * 10) +
            (dcftab[38] * 20) +
            (dcftab[39] * 40) +
            (dcftab[40] * 80);                                        // und Jahr speichern

          zeroflg = 0;                                                // Nullbyte-Flag lschen
          for (i = 0; i < 7; i ++) {                                  // 7 Puffer-Bytes vergleichen
            zeroflg = zeroflg | dcfbuf1[i];                           // Puffer 1 mit Flag verknpfen
            if (dcfbuf1[i] != dcfbuf2[i]) errcnt ++;                  // Vergleich Puffer 1 mit Puffer 2
            if (dcfbuf1[i] != dcfbuf3[i]) errcnt ++;                  // Vergleich Puffer 1 mit Puffer 3
            dcfbuf3[i] = dcfbuf2[i];                                  // Kopieren Puffer 2 -> Puffer 3
            dcfbuf2[i] = dcfbuf1[i];                                  // Kopieren Puffer 1 -> Puffer 2
          }
          if (zeroflg == 0) errcnt ++;                                // wenn alle Puffer-Bytes = 0 -> Fehler

          dcfbuf2[1] ++;                                              // Puffer 2 um eine Minute erhhen
          if (dcfbuf2[1] > 59) {                                      // wenn berlauf
            dcfbuf2[1] = 0;                                           // Minuten zurcksetzen
            dcfbuf2[2] ++;                                            // Stunde erhhen
          }
          dcfbuf3[1] ++;                                              // Puffer 3 um eine Minute erhhen
          if (dcfbuf3[1] > 59) {                                      // wenn berlauf
            dcfbuf3[1] = 0;                                           // Minuten zurcksetzen
            dcfbuf3[2] ++;                                            // Stunde erhhen
          }
        }

        if (errcnt == 0) {                                            // wenn gesamte Prozedur fehlerfrei -> Uhr stellen
          cli();                                                      // Interrupts sperren
          if (timezone != dcfbuf1[0]) {                               // wenn sich die Zeitzone gendert hat
            timezone = dcfbuf1[0];                                    // neue Zeitzone setzen
            saveflag = 1;                                             // genderte Einstellungen im EEPROM speichern
          }
          minute = dcfbuf1[1];                                        // Minute setzen
          hour = dcfbuf1[2];                                          // Stunde setzen
          day = dcfbuf1[3];                                           // Kalendertag setzen
          wday = dcfbuf1[4];                                          // Wochentag setzen
          month = dcfbuf1[5];                                         // Monat setzen
          year = dcfbuf1[6];                                          // Jahr setzen
          intcnt = 0;                                                 // Interrupt-Zhler (20ms) zurcksetzen
          second = 0;                                                 // Sekunden zurcksetzen
          sei();                                                      // Interrupts wieder freigeben
          syncstat = 3;                                               // Uhr ist jetzt mit DCF77 synchronisiert
          synctout = c_sync;                                          // Synchronisierungs-Timeout neu starten
          if (minute != 0) lasthour = hour;                           // wenn keine volle Stunde -> Stunde fr automatischen Farbwechsel speichern
          dcftout = 255;                                              // DCF-Synchronisierungs-Timeout-Zhler stoppen
          matrix_on();                                                // LED-Matrix wieder einschalten
        }
      }
    }
                                                                      // --- Taster auswerten ---
    if (key1stat && !key1quit) {                                      // wenn Taster 1 gedrckt und noch nicht quittiert
      key1quit = 1;                                                   // Taster 1 quittieren
      if (matrix_state() == 0) {                                      // wenn LED-Matrix ausgeschaltet
        matrix_on();                                                  // LED-Matrix wieder einschalten
        if (nighttime && nacttout == 255) nacttout = c_natout;        // wenn Nachtzeit und Timeout nicht aktiv -> Nachtaktivierungs-Timeout starten
      }
      else {                                                          // wenn LED-Matrix eingeschaltet
        if (nighttime) nacttout = c_natout;                           // wenn Nachtzeit -> Nachtaktivierungs-Timeout neu starten
        disptout = c_setup;                                           // Anzeige-Timeout-Zhler setzen
        if (dispmode != 4) {                                          // wenn Anzeige nicht im Einstellungs-Modus
          dispmode = 4;                                               // Anzeige in Einstellungs-Modus setzen
          setcount = 0;                                               // Einstellungszhler auf erste Einstellung setzen
          setmode = pgm_read_byte(&modelist[setcount]);               // Einstellung aus Tabelle holen
        }
        else {                                                        // wenn Anzeige im Einstellungs-Modus
          switch (setmode) {                                          // weiterer Verlauf vom Einstellungs-Modus abhngig
            case 24 ... 31:                                           // wenn Einstellung 24-31 (Sensorname)
              if (!editmode) {                                        // wenn nicht im Editier-Modus
                setcount ++;                                          // nchste Einstellung auswhlen
                setmode = pgm_read_byte(&modelist[setcount]);         // Einstellung aus Tabelle holen
              }
              else {                                                  // wenn im Editier-Modus
                editpos ++;                                           // auf nchste Zeichenposition setzen
                if (editpos > 4) {                                    // wenn letzte Position berschritten
                  editmode = 0;                                       // Editier-Modus ausschalten
                  editpos = 0;                                        // Editier-Position auf Anfang
                  setcount ++;                                        // nchste Einstellung auswhlen
                  setmode = pgm_read_byte(&modelist[setcount]);       // Einstellung aus Tabelle holen
                }
              }
              break;
            case 33 ... 36:                                           // wenn Einstellung 33-36 (Alarmname)
              if (!editmode) {                                        // wenn nicht im Editier-Modus
                setcount ++;                                          // nchste Einstellung auswhlen
                setmode = pgm_read_byte(&modelist[setcount]);         // Einstellung aus Tabelle holen
              }
              else {                                                  // wenn im Editier-Modus
                editpos ++;                                           // auf nchste Zeichenposition setzen
                if (editpos > 11) {                                   // wenn letzte Position berschritten
                  editmode = 0;                                       // Editier-Modus ausschalten
                  setcount ++;                                        // nchste Einstellung auswhlen
                  setmode = pgm_read_byte(&modelist[setcount]);       // Einstellung aus Tabelle holen
                }
              }
              break;
            default:                                                  // bei allen anderen Einstellungs-Modi
              setcount ++;                                            // nchste Einstellung auswhlen
              setmode = pgm_read_byte(&modelist[setcount]);           // Einstellung aus Tabelle holen
          }
          if (setcount > sizeof(modelist) - 1) {                      // wenn letzte Einstellung durchgefhrt
            dispmode = 0;                                             // Anzeigemodus auf Normal setzen
            if (alm_flag) dispmode = 1;                               // wenn Alarme aktiv -> Anzeigemodus auf Alarm setzen
            if (msg_flag) dispmode = 2;                               // wenn Nachricht aktiv -> Anzeigemodus auf Nachricht setzen
            editmode = 0;                                             // Editier-Modus ausschalten
            editpos = 0;                                              // Editier-Position auf Anfang
            matrtest = 0;                                             // LED-Matrix-Test-Modus ausschalten
            saveflag = 1;                                             // alle genderten Einstellungen im EEPROM speichern
            check_date();                                             // gegebenenfalls Tag korrigieren
          }
        }
      }
    }

    if (key1long == c_keyl1) {                                        // wenn Taster 1 lange gedrckt
      key1long = c_keyl1 - c_keyl2;                                   // Zhler auf Auto-Repeat-Wert einstellen
      if (dispmode == 4) {                                            // wenn Einstellungs-Modus aktiv
        dispmode = 0;                                                 // Anzeigemodus auf Normal setzen
        if (alm_flag) dispmode = 1;                                   // wenn Alarme aktiv -> Anzeigemodus auf Alarm setzen
        if (msg_flag) dispmode = 2;                                   // wenn Nachricht aktiv -> Anzeigemodus auf Nachricht setzen
        editmode = 0;                                                 // Editier-Modus ausschalten
        editpos = 0;                                                  // Editier-Position auf Anfang
        if (datacoun < 0) datacoun = 0;                               // wenn noch in Systemstart-Modus -> Datenfeldzhler auf Datum setzen
        matrtest = 0;                                                 // LED-Matrix-Test-Modus ausschalten
        saveflag = 1;                                                 // alle genderten Einstellungen im EEPROM speichern
        check_date();                                                 // gegebenenfalls Tag korrigieren
      }
    }

    if (key2stat && !key2quit) {                                      // wenn Taster 2 gedrckt und noch nicht quittiert
      key2quit = 1;                                                   // Taster 2 quittieren
      if (matrix_state() == 0) {                                      // wenn LED-Matrix ausgeschaltet
        matrix_on();                                                  // LED-Matrix wieder einschalten
        if (nighttime && nacttout == 255) nacttout = c_natout;        // wenn Nachtzeit und Timeout nicht aktiv -> Nachtaktivierungs-Timeout starten
      }
      else {                                                          // sonst
        if (nighttime) nacttout = c_natout;                           // wenn Nachtzeit -> Nachtaktivierungs-Timeout neu starten
        if (dispmode == 4) {                                          // wenn Einstellungs-Modus aktiv
          disptout = c_setup;                                         // Anzeige-Timeout-Zhler setzen
          switch (setmode) {                                          // weiterer Verlauf ist von der Einstellung abhngig
            case 0:                                                   // wenn Einstellung 0 (Farbe fr Uhrzeit)
              colclk ++;                                              // Farbe fr Uhrzeit erhhen
              if (colclk > 16) colclk = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 1:                                                   // wenn Einstellung 1 (Farbe fr Datum)
              coldat ++;                                              // Farbe fr Datum erhhen
              if (coldat > 16) coldat = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 2:                                                   // wenn Einstellung 2 (Farbe fr Alarme)
              colalm ++;                                              // Farbe fr Alarme erhhen
              if (colalm > 16) colalm = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 3:                                                   // wenn Einstellung 3 (Farbe meiden)
              avocol ++;                                              // zu meidende Farbe erhhen
              if (avocol > 16) avocol = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 4:                                                   // wenn Einstellung 4 (Minimale Helligkeit)
              brimin ++;                                              // Helligkeitswert erhhen
              if (brimin > 9) brimin = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 5:                                                   // wenn Einstellung 5 (Maximale Helligkeit)
              brimax ++;                                              // Helligkeitswert erhhen
              if (brimax > 9) brimax = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 6:                                                   // wenn Einstellung 6 (Test-LED)
              tstled ++;                                              // Test-LED erhhen
              if (tstled > 2) tstled = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 7:                                                   // wenn Einstellung 7 (Feste DCF-Synchronisierungszeit)
              dcfsyn ++;                                              // Feste DCF-Synchronisierungszeit erhhen
              if (dcfsyn > 1) dcfsyn = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              if (nighfr != nighto) dcfsyn = 0;                       // wenn Nachtzeit definiert -> Wert zurcksetzen
              break;
            case 8:                                                   // wenn Einstellung 8 (DCF-Invertierung)
              dcfinv ++;                                              // DCF-Invertierung erhhen
              if (dcfinv > 1) dcfinv = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 9:                                                   // wenn Einstellung 9 (DCF-Pull-up)
              dcfpup ++;                                              // DCF-Pull-up erhhen
              if (dcfpup > 1) dcfpup = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              if (dcfpup == 0) PORTA &= ~_BV(PINA6);                  // wenn Pull-up = 0 -> Pull-up am DCF-Eingang ausschalten
              else PORTA |= _BV(PINA6);                               // sonst Pull-up am DCF-Eingang einschalten
              break;
            case 10:                                                  // wenn Einstellung 10 (Stundengong)
              housnd ++;                                              // Stundengong erhhen
              if (housnd > 1) housnd = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 11:                                                  // wenn Einstellung 11 (Alarmsound)
              almsnd ++;                                              // Alarmsound erhhen
              if (almsnd > 1) almsnd = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 12:                                                  // wenn Einstellung 12 (Sound von)
              sounfr ++;                                              // Stundenwert erhhen
              if (sounfr > 23) sounfr = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 13:                                                  // wenn Einstellung 13 (Sound bis)
              sounto ++;                                              // Stundenwert erhhen
              if (sounto > 23) sounto = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 14:                                                  // wenn Einstellung 14 (Roll-Geschwindigkeit)
              msgscr ++;                                              // Wert erhhen
              if (msgscr > 20) msgscr = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 15:                                                  // wenn Einstellung 15 (Sensorfilter)
              senfil ++;                                              // Sensorfilter erhhen
              if (senfil > 1) senfil = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 16 ... 23:                                           // wenn Einstellung 16-23 (Sensornummer)
              sennum[setmode - 16] ++;                                // Sensornummer erhhen
              if (sennum[setmode - 16] > 31)                          // wenn Bereich berschritten
                sennum[setmode - 16] = 0;                             // Wert zurcksetzen
              break;
            case 24 ... 31:                                           // wenn Einstellung 24-31 (Sensorname)
              if (!editmode) {                                        // wenn nicht im Editier-Modus
                editmode = 1;                                         // Editier-Modus einschalten
                editpos = 0;                                          // Zeichenposition auf erstes Zeichen setzen
              }
              else {                                                  // wenn im Editier-Modus
                sennam[setmode - 24][editpos] =
                  next_char(sennam[setmode - 24][editpos]);           // nchstes Zeichen aus der Liste holen
              }
              break;
            case 32:                                                  // wenn Einstellung 32 (Alarmauswertung)
              alarms ++;                                              // Alarmauswertung erhhen
              if (alarms > 1) alarms = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              if (alarms == 1) {                                      // wenn Alarme eingeschaltet sind
                for (i = 0; i < 4; i ++) almnor[i] = almsnew[i];      // aktuellen Zustand der 4 Alarme als Normalzustand festlegen
              }
              else {                                                  // wenn Alarme ausgeschaltet sind
                for (i = 0; i < 4; i ++) {                            // 4 Alarme bearbeiten
                  almarray[i][0] = 0;                                 // Alarmfelder lschen
                  almstat[i] = 0;                                     // Alarmstatus lschen
                }
              }
              break;
            case 33 ... 36:                                           // wenn Einstellung 33-36 (Alarmname)
              if (!editmode) {                                        // wenn nicht im Editier-Modus
                editmode = 1;                                         // Editier-Modus einschalten
                editpos = 0;                                          // Zeichenposition auf erstes Zeichen setzen
              }
              else {                                                  // wenn im Editier-Modus
                almnam[setmode - 33][editpos] =
                  next_char(almnam[setmode - 33][editpos]);           // nchstes Zeichen aus der Liste holen
              }
              break;
            case 37:                                                  // wenn Einstellung 37 (Tag)
              day ++;                                                 // Tag erhhen
              if (day > 31) day = 1;                                  // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 38:                                                  // wenn Einstellung 38 (Monat)
              month ++;                                               // Monat erhhen
              if (month > 12) month = 1;                              // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 39:                                                  // wenn Einstellung 39 (Jahr)
              year ++;                                                // Jahr erhhen
              if (year > 99 ) year = c_year;                          // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 40:                                                  // wenn Einstellung 40 (Stunde)
              hour ++;                                                // Stunde erhhen
              if (hour > 23) hour = 0;                                // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 41:                                                  // wenn Einstellung 41 (Minute)
              minute ++;                                              // Minute erhhen
              if (minute > 59) minute = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 42:                                                  // wenn Einstellung 42 (Sekunde)
              cli();
              intcnt = 0;                                             // Interrupt-Zhler auf 0 stellen
              second = 0;                                             // Sekunden auf 0 stellen
              syncstat = 1;                                           // Uhr wurde manuell gestellt
              sei();
              timezone = chk_summtime();                              // Zeitzone ermitteln
              lasthour = hour;                                        // Stunde fr automatischen Farbwechsel speichern
              break;
            case 43:                                                  // wenn Einstellung 43 (Matrix-Test)
              matrtest ++;                                            // LED-Matrix-Test-Modus erhhen
              if (matrtest > 1) matrtest = 0;                         // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 44:                                                  // wenn Einstellung 44 (Sound-Test)
              sndflags[sndtest] = 1;                                  // Sound-Flag fr aktuellen Sound setzen
              sndtest ++;                                             // Zhler auf nchsten Sound setzen
              if (sndtest > 5) sndtest = 0;                           // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 45:                                                  // wenn Einstellung 45 (Sprache)
              lang ++;                                                // Sprache erhhen
              if (lang > 1) lang = 0;                                 // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 46:                                                  // wenn Einstellung 46 (Datum anzeigen)
              shdate ++;                                              // Datum anzeigen erhhen
              if (shdate > 1) shdate = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 47:                                                  // wenn Einstellung 47 (Wochentag anzeigen)
              shwday ++;                                              // Wochentag anzeigen erhhen
              if (shwday > 1) shwday = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 48:                                                  // wenn Einstellung 48 (Wechselzeit)
              chtime ++;                                              // Wechselzeit erhhen
              if (chtime > 4) chtime = 2;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 49:                                                  // wenn Einstellung 49 (Automatische Sommerzeit)
              autost ++;                                              // Automatische Sommerzeit erhhen
              if (autost > 1) autost = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 50:                                                  // wenn Einstellung 50 (Taganzeige)
              daymod ++;                                              // Taganzeige erhhen
              if (daymod > 2) daymod = 0;                             // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 51:                                                  // wenn Einstellung 51 (DCF77-Synchronisierungsstunde)
              dcfhou ++;                                              // DCF77-Synchronisierungsstunde erhhen
              if (dcfhou > 23) dcfhou = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 52:                                                  // wenn Einstellung 52 (Nacht von)
              nighfr ++;                                              // Stundenwert erhhen
              dcfsyn = 0;                                             // feste DCF-Synchronisierungszeit ausschalten
              if (nighfr > 23) nighfr = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
            case 53:                                                  // wenn Einstellung 53 (Nacht bis)
              nighto ++;                                              // Stundenwert erhhen
              dcfsyn = 0;                                             // feste DCF-Synchronisierungszeit ausschalten
              if (nighto > 23) nighto = 0;                            // wenn Bereich berschritten -> Wert zurcksetzen
              break;
          }
        }
        if (dispmode < 3) {                                           // wenn normale, Alarm- oder Nachrichten-Anzeige aktiv
          mute ++;                                                    // Mute erhhen
          if (mute > 1) mute = 0;                                     // wenn Bereich berschritten -> Wert zurcksetzen
          if (mute == 1) mutetout = c_mute;                           // wenn Mute aktiv -> Stummschaltungs-Timeout neu starten
        }
      }
    }

    if (key2long == c_keyl1) {                                        // wenn Taster 2 lange gedrckt
      key2long = c_keyl1 - c_keyl2;                                   // Zhler auf Auto-Repeat-Wert einstellen
      if (dispmode == 4) {                                            // wenn Einstellungs-Modus aktiv
        disptout = c_setup;                                           // Anzeige-Timeout-Zhler setzen
        switch (setmode) {                                            // weiterer Verlauf ist von der Einstellung abhngig
          case 4:                                                     // wenn Einstellung 4 (Minimale Helligkeit)
            brimin ++;                                                // Helligkeit erhhen
            if (brimin > 9) brimin = 0;                               // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 5:                                                     // wenn Einstellung 5 (Maximale Helligkeit)
            brimax ++;                                                // Helligkeit erhhen
            if (brimax > 9) brimax = 0;                               // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 12:                                                    // wenn Einstellung 12 (Sound von)
            sounfr ++;                                                // Stundenwert erhhen
            if (sounfr > 23) sounfr = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 13:                                                    // wenn Einstellung 13 (Sound bis)
            sounto ++;                                                // Stundenwert erhhen
            if (sounto > 23) sounto = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 14:                                                    // wenn Einstellung 14 (Roll-Geschwindigkeit)
            msgscr ++;                                                // Rollwert erhhen
            if (msgscr > 20) msgscr = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 16 ... 23:                                             // wenn Einstellung 15-23 (Sensornummer)
            sennum[setmode - 16] ++;                                  // Sensornummer erhhen
            if (sennum[setmode - 16] > 31) sennum[setmode - 16] = 0;  // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 24 ... 31:                                             // wenn Einstellung 24-31 (Sensorname)
            if (editmode) {                                           // wenn im Editier-Modus
              sennam[setmode - 24][editpos] =
                next_char(sennam[setmode - 24][editpos]);             // nchstes Zeichen aus der Liste holen
            }
            break;
          case 33 ... 36:                                             // wenn Einstellung 33-36 (Alarmname)
            if (editmode) {                                           // wenn im Editier-Modus
              almnam[setmode - 33][editpos] =
                next_char(almnam[setmode - 33][editpos]);             // nchstes Zeichen aus der Liste holen
            }
            break;
          case 37:                                                    // wenn Einstellung 37 (Tag)
            day ++;                                                   // Tag erhhen
            if (day > 31) day = 1;                                    // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 38:                                                    // wenn Einstellung 38 (Monat)
            month ++;                                                 // Monat erhhen
            if (month > 12) month = 1;                                // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 39:                                                    // wenn Einstellung 39 (Jahr)
            year ++;                                                  // Jahr erhhen
            if (year > 99) year = c_year;                             // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 40:                                                    // wenn Einstellung 40 (Stunde)
            hour ++;                                                  // Stunde erhhen
            if (hour > 23) hour = 0;                                  // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 41:                                                    // wenn Einstellung 41 (Minute)
            minute ++;                                                // Minute erhhen
            if (minute > 59) minute = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 51:                                                    // wenn Einstellung 51 (DCF77-Synchronisierungsstunde)
            dcfhou ++;                                                // DCF77-Synchronisierungsstunde erhhen
            if (dcfhou > 23) dcfhou = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 52:                                                    // wenn Einstellung 52 (Nacht von)
            nighfr ++;                                                // Stundenwert erhhen
            dcfsyn = 0;                                               // feste DCF-Synchronisierungszeit ausschalten
            if (nighfr > 23) nighfr = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
          case 53:                                                    // wenn Einstellung 53 (Nacht bis)
            nighto ++;                                                // Stundenwert erhhen
            dcfsyn = 0;                                               // feste DCF-Synchronisierungszeit ausschalten
            if (nighto > 23) nighto = 0;                              // wenn Bereich berschritten -> Wert zurcksetzen
            break;
        }
      }
      if (dispmode == 0) {                                            // wenn Normal-Modus aktiv
        key2long = 0;                                                 // Zhler fr langen Tastendruck zurcksetzen
        key2lcnt ++;                                                  // Zykluszhler fr langen Tastendruck erhhen
        if (key2lcnt == c_keyl3) {                                    // wenn ntige Dauer erreicht ist
          mute = 0;                                                   // Mute wieder ausschalten
          send_str(49);                                               // Poweroff-Kommando an RasPi senden
        }
      }
    }

    if (key3stat && !key3quit) {                                      // wenn Taster 3 gedrckt und noch nicht quittiert
      key3quit = 1;                                                   // Taster 3 quittieren
      if (matrix_state() == 0) {                                      // wenn LED-Matrix ausgeschaltet
        matrix_on();                                                  // LED-Matrix wieder einschalten
        if (nighttime && nacttout == 255) nacttout = c_natout;        // wenn Nachtzeit und Timeout nicht aktiv -> Nachtaktivierungs-Timeout starten
      }
      else {                                                          // sonst (wenn Matrix eingeschaltet)
        if (nighttime) nacttout = c_natout;                           // wenn Nachtzeit -> Nachtaktivierungs-Timeout neu starten
        if (dispmode < 3) {                                           // wenn normale, Alarm- oder Nachrichten-Anzeige aktiv
          quitcnt = 0;                                                // Quittierungszhler starten fr Animation
          if (msg_flag) {                                             // wenn Nachrichten-Flag gesetzt
            msg_flag = 0;                                             // Nachrichten-Flag lschen
            msgpbend = 0;                                             // Position vom Nachrichtenende lschen
          }
          else {                                                      // wenn Nachrichten-Flag nicht gesetzt
            for (i = 0; i < 8; i ++) {                                // 8 Alarm-Eintrge durchsuchen
              if (almstat[i] == 0) almarray[i][0] = 0;                // wenn Alarmstatus aus -> Alarmeintrag lschen
            }
            x = 0;                                                    // Flag zunchst lschen
            for (i = 0; i < 8; i ++) {                                // 8 Alarm-Eintrge durchsuchen
              if (almarray[i][0] > 0) x = 1;                          // wenn Alarm-Eintrag gefunden -> Flag setzen
            }
            alm_flag = x;                                             // ermittelten Alarmstatus in Alarm-Flag bertragen
            if (alm_flag) {                                           // wenn noch aktive Alarme anstehen
              almcoun ++;                                             // Alarmfeldzhler erhhen
              i = almcoun;                                            // Zhler zwischenspeichern
              while (almcoun < 8) {                                   // wiederholen solange Zhler im zulssigen Bereich
                if (almarray[almcoun][0] > 0) break;                  // wenn belegtes Alarmfeld gefunden -> Abbruch
                almcoun ++;                                           // Alarmfeldzhler erhhen
              }
              if (almcoun > 7) {                                      // wenn alle Alarmfelder ausgegeben
                almcoun = 0;                                          // Zhler zurcksetzen und
                while (almcoun < i) {                                 // weiter bis zum letzten Zhlerstand suchen
                  if (almarray[almcoun][0] > 0) break;                // wenn belegtes Alarmfeld gefunden -> Abbruch
                  almcoun ++;                                         // Alarmfeldzhler erhhen
                }
              }
            }
          }
        }        
      }
    }
                                                                      // --- Datenkommunikation bearbeiten ---
    char rx_char;                                                     // Zwischenspeicher fr empfangenes Zeichen
    while (rxwrite != rxread) {                                       // solange Schreib- und Lese-Zeiger unterschiedlich sind
      rx_char = rxbuff[rxread ++];                                    // empfangenes Zeichen aus USART-Puffer holen und Lese-Zeiger erhhen
      if ((rxpos < sizeof(rxdata)) && (rx_char >= 32)) {              // wenn Pufferende noch nicht erreicht und kein Steuerzeichen
        rxdata[rxpos ++] = rx_char;                                   // empfangenes Zeichen in RX-Datenpuffer legen und Zeichenzhler erhhen
      }
      if ((rx_char == '\r') || (rx_char == '\n')) {                   // wenn empfangenes Zeichen Carriage Return (ASCII 13) oder New Line (ASCII 10)
        if ((rxdata[1] == ':') && (rxpos == 7)) sensor_sm();          // wenn Doppelpunkt an Position 2 und insgesamt 7 Zeichen im Datenpuffer (Sensorwert vom Sensormodul) -> bearbeiten
        if ((rxdata[1] == ':') && (rxpos == 3)) alarm_sm();           // wenn Doppelpunkt an Position 2 und insgesamt 3 Zeichen im Datenpuffer (Alarm Sensormodul) -> bearbeiten
        if ((rxdata[2] == ':') && (rxpos > 2)) {                      // wenn Doppelpunkt an Position 3 und mindestens 3 Zeichen im Datenpuffer (Datenpaket vom Web-Modul)
          if ((rxdata[0] >= '0') && (rxdata[0] <= '3')) sensor_wm();  // wenn erste Stelle Ziffer zwischen 0 und 3 (Sensorwert oder Geburtstag vom Web-Modul) -> bearbeiten
          if (rxdata[0] == 'A') alarm_wm();                           // wenn erste Stelle "A" (Alarm vom Web-Modul) -> bearbeiten
          if ((rxdata[0] == 'T') && (rxdata[1] == 'M')) time_wm();    // wenn Kennung "TM" (Zeit vom Web-Modul) -> bearbeiten
          if ((rxdata[0] == 'M') && (rxdata[1] == 'E')) msg_wm();     // wenn Kennung "ME" (Textnachricht vom Web-Modul) -> bearbeiten
        }
        rxpos = 0;                                                    // RX-Datenpuffer wieder lschen
        tstlcnt = 0;                                                  // Test-LED-Zhler starten (LED einschalten)
      }
    }
                                                                      // --- Timeouts und Sonstiges bearbeiten ---
    if (disptout == 0) {                                              // wenn Anzeige-Timeout-Zhler abgelaufen
      disptout = 255;                                                 // Timeout-Zhler stoppen
      if (dispmode == 4) {                                            // wenn Anzeige im Einstellungsmodus
        matrtest = 0;                                                 // LED-Matrix-Test-Modus ausschalten
        saveflag = 1;                                                 // alle genderten Einstellungen im EEPROM speichern
        check_date();                                                 // gegebenenfalls Tag korrigieren
      }
      dispmode = 0;                                                   // Anzeigemodus auf Normal setzen
      if (alm_flag) dispmode = 1;                                     // wenn Alarme aktiv -> Anzeigemodus auf Alarm setzen
      if (msg_flag) dispmode = 2;                                     // wenn Nachricht aktiv -> Anzeigemodus auf Nachricht setzen
      editmode = 0;                                                   // Editier-Modus ausschalten
      editpos = 0;                                                    // Editier-Position auf Anfang
    }

    if (dcftout == 0) {                                               // wenn DCF-Synchronisierungs-Timeout-Zhler abgelaufen
      dcftout = 255;                                                  // Timeout-Zhler stoppen
      matrix_on();                                                    // LED-Matrix wieder einschalten
    }

    if (dcfsyn) {                                                     // wenn feste DCF-Synchronisierungszeit eingeschaltet
      if ((dispmode < 2) && (hour == 0) && (minute == 2) &&
        (second == 0) && (syncstat == 0)) {                           // wenn Uhr noch nie synchronisiert, normale Anzeige und Zeit = 0:02 Uhr (2 min nach Systemstart)
        matrix_off();                                                 // LED-Matrix ausschalten
        dcftout = c_dcf;                                              // DCF-Synchronisierungs-Timeout setzen
      }
      if ((dispmode < 2) && (hour == dcfhou) && (minute == 0) &&
        (second == 0)) {                                              // wenn normale Anzeige und Zeit = dcfhou (Feste Synchronisierungszeit)
        matrix_off();                                                 // LED-Matrix ausschalten
        dcftout = c_dcf;                                              // DCF-Synchronisierungs-Timeout setzen
      }
    }

    if (synctout == 0) {                                              // wenn Synchronisierungs-Timeout-Zhler abgelaufen
      synctout = 255;                                                 // Timeout-Zhler stoppen
      if (syncstat > 2) syncstat = 2;                                 // wenn Uhr synchronisiert war -> Status auf "Sync-Ausfall" setzen
    }

    if (mutetout == 0) {                                              // wenn Stummschaltungs-Timeout-Zhler abgelaufen
      mutetout = 255;                                                 // Timeout-Zhler stoppen
      mute = 0;                                                       // Stummschaltung ausschalten
    }

    if (nacttout == 0) nacttout = 255;                                // Wenn Nachtaktivierungs-Timeout-Zhler abgelaufen -> stoppen

    if ((hour == (lasthour + 1)) || (hour == (lasthour - 23))) {      // wenn regulrer Stundenwechsel
      lasthour = hour;                                                // neue Stunde speichern
      if (dispmode < 4) color_random();                               // wenn nicht im Einstellungs-Modus -> Farbwechsel durchfhren
    }

    if (dispmode < 3) {                                               // wenn normaler, Alarm- oder Nachrichten-Modus
      dispmode = 0;                                                   // Anzeigemodus zunchst auf Normal setzen
      if (alm_flag) dispmode = 1;                                     // wenn Alarme aktiv -> Anzeigemodus auf Alarm setzen
      if (msg_flag) dispmode = 2;                                     // wenn Nachricht aktiv -> Anzeigemodus auf Nachricht setzen
    }

    sounflag = 0;                                                     // Sound-Flag zunchst lschen
    if (sounfr < sounto) {                                            // wenn Beginnzeit < Endzeit
      if ((hour >= sounfr) && (hour <= sounto)) sounflag = 1;         // wenn aktuelle Zeit innerhalb der Beginn- und Endzeit -> Sound-Flag setzen
    }
    if (sounfr > sounto) {                                            // wenn Beginnzeit > Endzeit
      if ((hour >= sounfr) || (hour <= sounto)) sounflag = 1;         // wenn aktuelle Zeit nach Beginnzeit oder vor Endzeit -> Sound-Flag setzen
    }
    if ((hour == sounto) && (minute > 0)) sounflag = 0;               // wenn Endzeit erreicht und Minute 0 berschritten -> Sound-Flag lschen (Sound bleibt noch eine Minute lnger aktiv)

    if (syncstat && housnd && sounflag && !mute &&                    // wenn Uhr synchronisiert, Stundensound eingeschaltet, Soundzeit, keine Stummschaltung,
      (minute == 0) && (second == 0) && !sndflags[5])                 // Minuten = 0, Sekunden = 0 und Flag fr Stundensound nicht gesetzt
        sndflags[5] = 1;                                              // Stundensound auslsen

    if (syncstat > 0) {                                               // wenn Uhr synchronisiert
      nighttime = 0;                                                  // Nacht-Zeit-Flag zunchst lschen
      if (nighfr < nighto) {                                          // wenn Beginnzeit < Endzeit
        if (hour >= nighfr && hour < nighto) nighttime = 1;           // wenn aktuelle Zeit innerhalb der Beginn- und Endzeit -> Nacht-Zeit-Flag setzen
      }
      if (nighfr > nighto) {                                          // wenn Beginnzeit > Endzeit
        if (hour >= nighfr || hour < nighto) nighttime = 1;           // wenn aktuelle Zeit nach Beginnzeit oder vor Endzeit -> Nacht-Zeit-Flag setzen
      }
    }
    else nighttime = 0;                                               // wenn Uhr nicht synchronisiert -> Nacht-Zeit-Flag lschen
    if (dispmode < 4 && nighttime && nacttout == 255) {               // wenn normale Anzeige, Nacht-Zeit und Nachtaktivierungs-Zhler inaktiv
      if (matrix_state() > 0) matrix_off();                           // wenn LED-Matrix eingeschaltet -> ausschalten
    }
    else {                                                            // wenn Nacht-Zeit und Nachtaktivierungs-Zhler aktiv
      if (matrix_state() == 0) matrix_on();                           // wenn LED-Matrix ausgeschaltet -> einschalten
    }

    x = 0;                                                            // Summenwert fr Helligkeit lschen
    for (i = 0; i < 32; i ++) x += bright[i];                         // alle Werte aus dem Helligkeits-Puffer addieren
    i = x / 78;                                                       // Summenwert (maximal 255*32 = 8160) in PWM-Wert 0-105 fr Helligkeitssteuerung umrechnen
    j = pgm_read_byte(&brimintab[brimin]);                            // Wert fr Minimale Helligkeit aus Tabelle holen
    if (i < j) i = j;                                                 // wenn Minimale Helligkeit unterschritten -> auf eingestellten Minimalwert erhhen
    j = pgm_read_byte(&brimaxtab[brimax]);                            // Wert fr Maximale Helligkeit aus Tabelle holen
    if (i > j) i = j;                                                 // wenn Maximale Helligkeit berschritten -> auf eingestellten Maximalwert vermindern
    bripwm = i;                                                       // ermittelten PWM-Wert setzen

    if (saveflag) {                                                   // wenn Einstellungen im EEPROM gespeichert werden mssen
      saveflag = 0;                                                   // Flag wieder lschen
      if (lang != eep_read_byte(e_lang))                              // wenn nderung
        eep_write_byte(e_lang, lang);                                 // Sprache ins EEPROM schreiben
      if (dcfsyn != eep_read_byte(e_dcfsyn))                          // wenn nderung
        eep_write_byte(e_dcfsyn, dcfsyn);                             // Feste DCF77-Synchronisierungszeit ins EEPROM schreiben
      if (dcfinv != eep_read_byte(e_dcfinv))                          // wenn nderung
        eep_write_byte(e_dcfinv, dcfinv);                             // DCF77-Invertierung ins EEPROM schreiben
      if (dcfpup != eep_read_byte(e_dcfpup))                          // wenn nderung
        eep_write_byte(e_dcfpup, dcfpup);                             // DCF77-Pull-up ins EEPROM schreiben
      if (tstled != eep_read_byte(e_tstled))                          // wenn nderung
        eep_write_byte(e_tstled, tstled);                             // Test-LED ins EEPROM schreiben
      if (colclk != eep_read_byte(e_colclk))                          // wenn nderung
        eep_write_byte(e_colclk, colclk);                             // Farbe fr Uhrzeit ins EEPROM schreiben
      if (coldat != eep_read_byte(e_coldat))                          // wenn nderung
        eep_write_byte(e_coldat, coldat);                             // Farbe fr Datum, Wochentag und Sensordaten ins EEPROM schreiben
      if (colalm != eep_read_byte(e_colalm))                          // wenn nderung
        eep_write_byte(e_colalm, colalm);                             // Farbe fr Alarmmeldungen ins EEPROM schreiben
      if (brimin != eep_read_byte(e_brimin))                          // wenn nderung
        eep_write_byte(e_brimin, brimin);                             // Minimale Helligkeit ins EEPROM schreiben
      if (brimax != eep_read_byte(e_brimax))                          // wenn nderung
        eep_write_byte(e_brimax, brimax);                             // Maximale Helligkeit ins EEPROM schreiben
      if (senfil != eep_read_byte(e_senfil))                          // wenn nderung
        eep_write_byte(e_senfil, senfil);                             // Sensorfilter ins EEPROM schreiben
      if (alarms != eep_read_byte(e_alarms))                          // wenn nderung
        eep_write_byte(e_alarms, alarms);                             // Alarmauswertung ins EEPROM schreiben
      for (i = 0; i < 8; i ++) {                                      // 8 Sensornummern bearbeiten
        if (sennum[i] != eep_read_byte(e_sennum + i))                 // wenn nderung
          eep_write_byte(e_sennum + i, sennum[i]);                    // Sensornummer ins EEPROM schreiben
      }
      for (i = 0; i < 8; i ++) {                                      // 8 Sensornamen bearbeiten
        for (j = 0; j < 5; j ++) {                                    // jeweils 5 Zeichen bearbeiten
          if (sennam[i][j] != eep_read_byte(e_sennam + i * 5 + j))    // wenn nderung
            eep_write_byte(e_sennam + i * 5 + j, sennam[i][j]);       // Zeichen ins EEPROM schreiben
        }
      }
      for (i = 0; i < 4; i ++) {                                      // 4 Alarmzustnde bearbeiten
        if (almnor[i] != eep_read_byte(e_almnor + i))                 // wenn nderung
          eep_write_byte(e_almnor + i, almnor[i]);                    // Alarmzustand ins EEPROM schreiben
      }
      for (i = 0; i < 4; i ++) {                                      // 4 Alarmnamen bearbeiten
        for (j = 0; j < 12; j ++) {                                   // jeweils 12 Zeichen bearbeiten
          if (almnam[i][j] != eep_read_byte(e_almnam + i * 12 + j))   // wenn nderung
            eep_write_byte(e_almnam + i * 12 + j, almnam[i][j]);      // Zeichen ins EEPROM schreiben
        }
      }
      if (housnd != eep_read_byte(e_housnd))                          // wenn nderung
        eep_write_byte(e_housnd, housnd);                             // Stundengong ins EEPROM schreiben
      if (almsnd != eep_read_byte(e_almsnd))                          // wenn nderung
        eep_write_byte(e_almsnd, almsnd);                             // Alarmsound ins EEPROM schreiben
      if (sounfr != eep_read_byte(e_sounfr))                          // wenn nderung
        eep_write_byte(e_sounfr, sounfr);                             // Sound von ins EEPROM schreiben
      if (sounto != eep_read_byte(e_sounto))                          // wenn nderung
        eep_write_byte(e_sounto, sounto);                             // Sound bis ins EEPROM schreiben
      if (msgscr != eep_read_byte(e_msgscr))                          // wenn nderung
        eep_write_byte(e_msgscr, msgscr);                             // Roll-Geschwindigkeit ins EERPOM schreiben
      if (avocol != eep_read_byte(e_avocol))                          // wenn nderung
        eep_write_byte(e_avocol, avocol);                             // Farbe meiden ins EEPROM schreiben
      if (shdate != eep_read_byte(e_shdate))                          // wenn nderung
        eep_write_byte(e_shdate, shdate);                             // Datum anzeigen ins EEPROM schreiben
      if (shwday != eep_read_byte(e_shwday))                          // wenn nderung
        eep_write_byte(e_shwday, shwday);                             // Wochentag anzeigen ins EEPROM schreiben
      if (chtime != eep_read_byte(e_chtime))                          // wenn nderung
        eep_write_byte(e_chtime, chtime);                             // Wechselzeit ins EEPROM schreiben
      if (autost != eep_read_byte(e_autost))                          // wenn nderung
        eep_write_byte(e_autost, autost);                             // Automatische Sommerzeit ins EEPROM schreiben
      if (timezone != eep_read_byte(e_timezo))                        // wenn nderung
        eep_write_byte(e_timezo, timezone);                           // Zeitzone ins EEPROM schreiben
      if (daymod != eep_read_byte(e_daymod))                          // wenn nderung
        eep_write_byte(e_daymod, daymod);                             // Taganzeige-Modus ins EEPROM schreiben
      if (dcfhou != eep_read_byte(e_dcfhou))                          // wenn nderung
        eep_write_byte(e_dcfhou, dcfhou);                             // DCF77-Synchronisierungsstunde ins EEPROM schreiben
      if (nighfr != eep_read_byte(e_nighfr))                          // wenn nderung
        eep_write_byte(e_nighfr, nighfr);                             // Nacht von ins EEPROM schreiben
      if (nighto != eep_read_byte(e_nighto))                          // wenn nderung
        eep_write_byte(e_nighto, nighto);                             // Nacht bis ins EEPROM schreiben
    }
  }
}
